# BELONG TO SERVICE TASK
import os, re, json, time, uuid, logging, pytz, socket, traceback
from enum import Enum
from datetime import datetime
from math import inf as positive_infinity
from pymongo import MongoClient
from psycopg2.extras import RealDictCursor
from enums import OperatingSystem, SupportedOS, DBType, ClusterRoles, RestoreOptions, Status, Severity, Inventory, Schedule, LicenseType, Repository, Timeout
from tasks_core import insert_task, remote_command, check_remote_directory_exists, create_ssh_connection, get_postgres_connection, logger, \
    check_os_version, find_tilda, get_date_modified_folder, update_failed_object, install_offline_remote_to_remote, generate_complex_password, \
    TASK_TMP_FILES_DIR, TASK_CERTIFICATE_FILES_DIR, transfer_files_rsync, write_to_local_file, transfer_ssl_certificates, init_container, get_log_rotate_config, local_command, check_local_directory_exists, find_local_tilda

class ReadPreference(Enum):
    PRIMARY_PREFERRED = 'primaryPreferred'
    PRIMARY = 'primary'
    SECONDARY_PREFERRED = 'secondaryPreferred'
    SECONDARY = 'secondary'
    NEAREST = 'nearest'

#####################################################################
# MONGODB TASKS
#####################################################################

def create_mongo_connection(node, hostname, read_preference = ReadPreference.PRIMARY.value,  direct = False, no_auth = False, no_ssl = False, mongos_port = None):
    """get connection to mongo node via pymongo"""
    try:
        use_ssl = node['ex_params'].get('ssl')
        # assumes mongos are on same host as config replicaset - to be changed if otherwise
        connect_port = int(mongos_port if mongos_port else node['port'])
        hostname =  hostname.strip()
        kwargs = {
            'readPreference': read_preference,
            'directConnection': True if ClusterRoles.ARBITER.value in node['cluster_role'] else direct,
            'serverSelectionTimeoutMS': Timeout.SERVICE_MONGO.value
        }
        
        if not no_auth:
            kwargs['username'] = node['database_user'] 
            kwargs['password'] = node['database_pwd'] 
        
        if use_ssl and not no_ssl:
            cert_root = f"{TASK_CERTIFICATE_FILES_DIR}/{hostname}_{node['port']}"
            ca = f"{cert_root}/ca.crt"
            key = f"{cert_root}/ca.pem"
            kwargs['tls'] = True
            kwargs['tlsCAFile'] = ca
            kwargs['tlsCertificateKeyFile'] = key
            kwargs['tlsAllowInvalidHostnames'] = True
        return MongoClient(hostname, connect_port, **kwargs)
    except Exception as e:
        return (False, e)

def write_ssl_to_file(host, cert_root, mongo_port=27017):
    """write a local copy of node ssl certificates for usage in pymongo connections of task_service"""

    try:
        ssl_backup_path = f"{TASK_CERTIFICATE_FILES_DIR}/{host['name'].strip()}_{mongo_port}"
        ca_crt = f'{cert_root}/ca.crt'
        ca_pem = f'{cert_root}/ca.pem'
        # just directly fetch by python ssh to RAM, 
        # since node host might not have sftp server configured and task service os might not have built-in scp/rsync
        jsn = json.loads(remote_command(host, f"sudo cat '{ca_pem}'", True), strict=False)
        if jsn['stderr'] != '':
            return jsn

        jsn = write_to_local_file(f"{ssl_backup_path}/{os.path.basename(ca_pem)}", jsn['stdout'])
        if jsn['stderr'] != '':
            return jsn

        jsn = json.loads(remote_command(host, f"sudo cat '{ca_crt}'", True), strict=False)
        if jsn['stderr'] != '':
            return jsn

        jsn = write_to_local_file(f"{ssl_backup_path}/{os.path.basename(ca_crt)}", jsn['stdout'])
        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}

def setup_mongo_ssl(host, node, ssl_backup_host=None)-> dict:
    """configure instance to support ssl connections
       ssl_backup_host - host to backup node ssl files into
    """

    try:
        jsn = generate_ssl_certificates(host, node, ssl_backup_host = ssl_backup_host)
        if jsn['stderr'] != '':
            # if fail to generate certificates - interrupt immediately
            return jsn

        # add tls section to node config file
        jsn = inject_ssl_config(host, node['dir_123Cluster'])
        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}

def generate_ssl_certificates(host, node, days=3650, ssl_backup_host=None) -> dict:
    """generate certificates for ssl connections support\n
       ssl_backup_host - host to backup node ssl files into
    """
    # TODO: -subj

    try:
        # we do plain rsync data folder backups, certificates should be at least unique per host
        # so for starters - common certificate location are data subfolder of node
        cert_root = f"{node['dir_123Cluster'].strip()}/data"
        # for compatibility (because may be ip), if use dns - name from /etc/hostname are preferred
        hostname = host['name']
        node_name = node['name'].strip()
        # sometimes (on centos 9, for example) - we could be prohibited from write into /tmp, better use path under ours control
        # (and include host_port to know which target are failed if something go wrong)
        conf_file = f'{TASK_TMP_FILES_DIR}/{node_name}/123cluster_ext.txt'
        # dont care if its ip or not, subjectAltName are reads only if hostname in certificate common name does not match contents of machine /etc/hostname file
        config = f'''[ req ]
default_bits       = 2048
prompt             = no
encrypt_key        = no
default_md         = sha256
x509_extensions    = req_ext
distinguished_name = dn

[ dn ]
C  = IL
ST = Test
L  = Tel-Aviv
O  = 123Cluster
OU = IT
emailAddress = admin@123Cluster.com
CN = {hostname}

[ req_ext ]
subjectAltName = @alt_names

[ alt_names ]
IP.1 = {hostname}'''
        ca_key = f'{cert_root}/ca.key'
        ca_crt = f'{cert_root}/ca.crt'
        ca_pem = f'{cert_root}/ca.pem'

        commands = [
            f"sudo mkdir -p '{os.path.dirname(conf_file)}'",
            f"echo '{config}' | sudo tee '{conf_file}'",
            f"sudo openssl req -new -x509  -days {days}  -config '{conf_file}' -out '{ca_crt}' -keyout '{ca_key}'",
            f"sudo cat '{ca_key}' '{ca_crt}' | sudo tee '{ca_pem}'",
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                return jsn

        jsn = set_ssl_permissions(host, node['dir_123Cluster'])
        if jsn['stderr'] != '':
            return jsn

        if ssl_backup_host:
            jsn = transfer_ssl_certificates(node, host, ssl_backup_host)
            if jsn['stderr'] != '':
                return jsn

            ssl_hostname = ssl_backup_host['name'].strip()
            task_service_hostname = socket.getfqdn()
            task_service_ip = socket.gethostbyname(task_service_hostname)
            if ssl_hostname not in task_service_hostname and ssl_hostname not in task_service_ip:
                jsn = write_ssl_to_file(host, cert_root, node['port'])

        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}

def inject_ssl_config(host, node_dir, mongos=False) -> dict:
    """insert tls section into mongo net config"""
    try:
        cert_root = f"{node_dir}/data"
        ca_crt = f'{cert_root}/ca.crt'
        ca_pem = f'{cert_root}/ca.pem'
        unit = 'mongos' if mongos else 'mongod'
        tls_config = f'''    tls:
        mode: requireTLS
        certificateKeyFile: {ca_pem}
        CAFile: {ca_crt}
        allowInvalidHostnames: true'''
        config_file = f'{node_dir}/conf/{unit}.cfg'
        jsn = json.loads(remote_command(host, f"sudo cat '{config_file}'", True), strict=False)
        if jsn['stderr'] != '':
            return jsn
        
        config = jsn['stdout']
        config = re.sub(r'(?<=\n  tls:\n)(    .*\n)*', '', config)
        config = re.sub(r'(  bindIp:.*)', rf'\1\n{tls_config}', config)
        jsn = json.loads(remote_command(host, f"echo '{config}' | sudo tee '{config_file}'", True), strict=False)
        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}

def set_ssl_permissions(host, node_dir) -> dict:
    """setup right permissions on mongo ssl files"""
    try:
        cert_root = f"{node_dir}/data"
        ca_crt = f'{cert_root}/ca.crt'
        ca_pem = f'{cert_root}/ca.pem'
        commands = [
            f"sudo chmod 400 '{ca_crt}'",
            f"sudo chmod 400 '{ca_pem}'"
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                return jsn

        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}
    
def check_mongo_version_exists(host, mongo_version, distributor) -> bool:
    cmd = f"rpm -qa | grep mongodb-org-server-{mongo_version}" if OperatingSystem.CENTOS.value in distributor else f"dpkg -l | grep mongodb-org-server | grep {mongo_version}"
    jsn = json.loads(remote_command(host, cmd, True), strict=False)
    return jsn['stdout'] != ''

def find_mongo_bin(host, mongo_version) -> str:
    """find location of mongo bin"""

    mongo_bin = '/usr/bin'
    try:
        jsn = json.loads(remote_command(host, f'sudo dirname $(rpm -ql mongodb-org-server-{mongo_version}*|grep /bin/mongod|grep -v share)', True), strict=False)
        if jsn['stderr'] != '':
            if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                jsn = json.loads(remote_command(host, 'sudo dirname $(which mongod)', True), strict=False)
                if jsn['stderr'] != '':
                    if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                        return mongo_bin

            return mongo_bin
        elif jsn['stdout'] != '':
            return jsn['stdout'].strip()
    except Exception as error:
        logging.error(f"find_mongo_bin: {str(error)}")

    return mongo_bin


def find_local_mongo_bin(mongo_version) -> str:
    """find location of mongo bin"""

    mongo_bin = '/usr/bin'
    try:
        jsn = json.loads(local_command(f'sudo dirname $(rpm -ql mongodb-org-server-{mongo_version}*|grep /bin/mongod|grep -v share)', True), strict=False)
        if jsn['stderr'] != '':
            if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                jsn = json.loads(local_command('sudo dirname $(which mongod)', True), strict=False)
                if jsn['stderr'] != '':
                    if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                        return mongo_bin

            return mongo_bin
        elif jsn['stdout'] != '':
            return jsn['stdout'].strip()
    except Exception as error:
        logging.error(f"find_mongo_bin: {str(error)}")

    return mongo_bin

def install_mongo_node_task(data, task_id):
    """
    A task that installs a single-node mongo replica set on a remote machine.
    MongoDB only supports Oracle Linux running the Red Hat Compatible Kernel (RHCK).
    MongoDB does not support the Unbreakable Enterprise Kernel (UEK).
    MongoDB only supports the 64-bit versions of RHEL / CentOS / Oracle 7/8.
    By default, MongoDB runs using the mongod user account and uses the following default directories:
        /var/lib/mongo (the data directory)
        /var/log/mongodb (the log directory)
    To use a data directory and/or log directory other than the default directories:
        Create the new directory or directories.
        Edit the configuration file /etc/mongod.conf and modify the following fields accordingly:
            storage.dbPath to specify a new data directory path (e.g. /some/data/directory)
            systemLog.path to specify a new log file path (e.g. /some/log/directory/mongod.log)
        Ensure that the user running MongoDB has access to the directory or directories:
            sudo chown -R mongod:mongod <directory>
    """

    task = 'install mongo node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        use_ssl = node['ex_params'].get('ssl')
        app_host = json.loads(data['app_host'])
        # app_hostname = data['app_hostname']
        create_run_linux = data['create_run_linux']
        mongo_version = float(data['software_version'])
        # repository = network or local
        repository = data['repository']
        org_id = node['org_id']
        host_exists = True
        proxmox_container = node['ex_params'].get('proxmox')
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = node['ex_params'].get('distributor')
                release = node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
          
        
        version_exists = check_mongo_version_exists(host, mongo_version, distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    if not version_exists:
        step = 'installing mongodb'
        duration_step_start = time.perf_counter()

        if Repository.LOCAL.value in repository:
            jsn = install_offline_remote_to_remote(distributor, release, host, mongo_version, DBType.MONGO.value, app_host)
        else:
            jsn = install_mongo_online(distributor, release, host, mongo_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating node'
    duration_step_start = time.perf_counter()

    jsn = setup_node_directory(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    jsn = start_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'configuring authentication'
    duration_step_start = time.perf_counter()

    error = create_superuser(host, node)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    data_path = node['dir_123Cluster']
    if data_path.startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        tilda_path = jsn['stdout']
        data_path = data_path.replace('~', tilda_path)

    jsn = stop_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    if use_ssl:
        jsn = setup_mongo_ssl(host, node, ssl_backup_host=app_host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
     
    #moved to here, authorization must be added only after the create superuser and configure ssl (and also to avoid editing config of running node)
    try:
        config_content = f"""
security:
    authorization: enabled
    keyFile: {data_path}/conf/key.file
"""
        if mongo_version >= 7.0:
            config_content += """
    clusterAuthMode: keyFile
"""
        remote_command(host, f"sudo openssl rand -base64 755 | sudo tee '{data_path}/conf/key.file'", True)
        remote_command(host, f"sudo chmod 400 '{data_path}/conf/key.file'", True)
        remote_command(host, f"echo '{config_content}' | sudo tee -a '{data_path}/conf/mongod.cfg'", True)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    jsn = start_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'initializing replicaset'
    duration_step_start = time.perf_counter()

    add_to_cfg(host, node)

    jsn = restart_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    error = add_first_node_to_replicaset(host, node)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    try:
        if mongo_version >= 4.4:
            # connection for primary
            client = create_mongo_connection(node, host['name'])
            if isinstance(client, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
                return

            with client:
                client.admin.command(({"setDefaultRWConcern" : 1, "defaultWriteConcern" : {"w" : 1}}))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024

        remote_command(host, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task', 'monitor_name': None, 'org_id': node['org_id'],   'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        result = insert_task(params)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['version'], node_size, node['id'],))
                conn.commit()  
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)

    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])

    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as err:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def install_mongos_node_task(data, task_id):
    """
    A task that installs a mongos router node for MongoDB sharded cluster on a remote machine.
    MongoDB only supports Oracle Linux running the Red Hat Compatible Kernel (RHCK).
    MongoDB does not support the Unbreakable Enterprise Kernel (UEK).
    MongoDB only supports the 64-bit versions of RHEL / CentOS / Oracle 7/8.
    By default, MongoDB runs using the mongod user account and uses the following default directories:
        /var/lib/mongo (the data directory)
        /var/log/mongodb (the log directory)
    To use a data directory and/or log directory other than the default directories:
        Create the new directory or directories.
        Edit the configuration file /etc/mongod.conf and modify the following fields accordingly:
            storage.dbPath to specify a new data directory path (e.g. /some/data/directory)
            systemLog.path to specify a new log file path (e.g. /some/log/directory/mongod.log)
        Ensure that the user running MongoDB has access to the directory or directories:
            sudo chown -R mongod:mongod <directory>
    """

    task = 'install mongos node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id, task_log_id, task, step, status, duration, created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id, task_log_id, task, step, status, duration, created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        use_ssl = node['ex_params'].get('ssl')
        app_host = json.loads(data['app_host'])
        # app_hostname = data['app_hostname']
        create_run_linux = data['create_run_linux']
        mongo_version = float(data['software_version'])
        # repository = network or local
        repository = data['repository']
        org_id = node['org_id']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,release['stdout'],release['stderr'],release['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,release['stdout'],release['stderr'],release['excode'])
            update_failed_object(node['id'])
            return
        
        version_exists = check_mongo_version_exists(host, mongo_version, distributor)

        data_path = node['dir_123Cluster']
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id, task_log_id, task, step, status, duration_step_end, created_by, None, str(error), 9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id, task_log_id, task, step, status, duration_task_end, created_by, None, str(error), 9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id, task_log_id, task, step, status, duration_step_end, created_by)

    if not version_exists:
        step = 'installing mongodb'
        duration_step_start = time.perf_counter()

        if Repository.LOCAL.value in repository:
            jsn = install_offline_remote_to_remote(distributor, release, host, mongo_version, DBType.MONGO.value, app_host)
        else:
            jsn = install_mongo_online(distributor, release, host, mongo_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id, task_log_id, task, step, status, duration_step_end, created_by)

    step = 'creating config standalone'
    duration_step_start = time.perf_counter()

    jsn = setup_node_directory(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    jsn = start_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return
    
    error = create_superuser(host, node, shard = True)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id, task_log_id, task, step, status, duration_step_end, created_by, None, error, 1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id, task_log_id, task, step, status, duration_task_end, created_by, None, error, 1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id, task_log_id, task, step, status, duration_step_end, created_by)
    
    step = 'initializing config replicaset'
    duration_step_start = time.perf_counter()
    
    add_to_cfg(host, node, config_rs=True)

    jsn = restart_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    error = add_first_node_to_replicaset(host, node, config_rs=True)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    try:
        if mongo_version >= 4.4:
            # connection for primary
            client = create_mongo_connection(node, host['name'], no_auth=True, no_ssl=True)
            if isinstance(client, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
                return

            with client:
                client.admin.command(({"setDefaultRWConcern" : 1, "defaultWriteConcern" : {"w" : 1}}))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
     
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id, task_log_id, task, step, status, duration_step_end, created_by)
           
    
    step = 'creating mongos router node'
    duration_step_start = time.perf_counter()
    
    # since we anyway creating only one mongos node - dont care about where it goes
    # (if need several nodes - rewrite)
    jsn = set_mongo_config_file(host, node, mongos = True)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    

    
    jsn = start_mongos(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return


    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id, task_log_id, task, step, status, duration_step_end, created_by)

    step = 'setting up sharded cluster authentication'
    duration_step_start = time.perf_counter()

    error = create_superuser(host, node, mongos = True)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id, task_log_id, task, step, status, duration_step_end, created_by, None, error, 1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id, task_log_id, task, step, status, duration_task_end, created_by, None, error, 1)
        return

    jsn = stop_mongo_process(host, node['ex_params'].get('mongos_port'))
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    jsn = stop_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    if use_ssl:
        jsn = setup_mongo_ssl(host, node, ssl_backup_host=app_host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return


      # must be in try-except block, if at least one command fail - primary will be dead for sure
    try:
        config_content = f"""\
security:
    authorization: enabled
    keyFile: {data_path}/conf/key.file
"""
        if mongo_version >= 7.0:
            config_content += """
    clusterAuthMode: keyFile
"""
        remote_command(host, f"sudo openssl rand -base64 755 | sudo tee '{data_path}/conf/key.file'", True)
        remote_command(host, f"sudo chmod 400 '{data_path}/conf/key.file'", True)
        remote_command(host, f"echo '{config_content}' | sudo tee -a '{data_path}/conf/mongod.cfg'", True)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id, task_log_id, task, step, status, duration_step_end, created_by, None, str(error), 1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id, task_log_id, task, step, status, duration_task_end, created_by, None, str(error), 1)    
    
    jsn = start_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    if use_ssl:
        jsn = inject_ssl_config(host, data_path, mongos = True)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    

    # must be in try-except block, if at least one command fail - primary will be dead for sure
    try:
        config_content = f"""\
security:
    keyFile: {data_path}/conf/key.file
"""
        if mongo_version >= 7.0:
            config_content += """
    clusterAuthMode: keyFile
"""
        remote_command(host, f"echo '{config_content}' | sudo tee -a '{data_path}/conf/mongos.cfg'", True)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id, task_log_id, task, step, status, duration_step_end, created_by, None, str(error), 1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id, task_log_id, task, step, status, duration_task_end, created_by, None, str(error), 1)   
    
    jsn = start_mongos(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id, task_log_id, task, step, status, duration_step_end, created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()
    
    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
        node_size = float(jsn.get('stdout').strip()) / 1024

        remote_command(host, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task', 'monitor_name': None,  'org_id': node['org_id'],  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role =%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.BALANCER.value, node_size, node['id'],))
                conn.commit()
                cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"severity":\"%s\"}' WHERE id=%s""" % (Severity.OK.value, node['cluster_id'],))
                conn.commit()
                cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"status":\"\"}' WHERE id=%s""", (node['cluster_id'],))
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id, node['id'], created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id, host['id'], created_by, OperatingSystem.LINUX.value])
                conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id, task_log_id, task, step, status, duration, created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id, task_log_id, task, step, status, duration_task_end, created_by)


def start_mongos(host, node) -> dict:
    """start a Mongos router node daemon"""

    try:

        mongo_version = float(node['version'])
        data_path = node['dir_123Cluster']
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        mongo_bin = find_mongo_bin(host, mongo_version) 
        command = f"""sudo {mongo_bin}/mongos --fork --port '{node['ex_params'].get('mongos_port')}' --config '{data_path}/conf/mongos.cfg'"""
        jsn = json.loads(remote_command(host, command, True), strict=False)
        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def install_mongo_shard_task(data, task_id):
    """
    A task that installs a single-node mongo shard replica set on a remote machine.
    MongoDB only supports Oracle Linux running the Red Hat Compatible Kernel (RHCK).
    MongoDB does not support the Unbreakable Enterprise Kernel (UEK).
    MongoDB only supports the 64-bit versions of RHEL / CentOS / Oracle 7/8.
    By default, MongoDB runs using the mongod user account and uses the following default directories:
        /var/lib/mongo (the data directory)
        /var/log/mongodb (the log directory)
    To use a data directory and/or log directory other than the default directories:
        Create the new directory or directories.
        Edit the configuration file /etc/mongod.conf and modify the following fields accordingly:
            storage.dbPath to specify a new data directory path (e.g. /some/data/directory)
            systemLog.path to specify a new log file path (e.g. /some/log/directory/mongod.log)
        Ensure that the user running MongoDB has access to the directory or directories:
            sudo chown -R mongod:mongod <directory>
    """

    task = 'install mongo shard task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        # combined data of config replicaset primary instance + mongos node,
        # because Evgeniy wrote in install_mongos_node_task config server primary to be on same host as mongos,
        # might be changed if we add features of adding new nodes to config replicaset/adding new mongos nodes 
        config_host = json.loads(data['mongos_host_data'])
        config_node = json.loads(data['mongos_node_data'])
        mongos_port = config_node['ex_params'].get('mongos_port')
        use_ssl = node['ex_params'].get('ssl')
        app_host = json.loads(data['app_host'])
        # app_hostname = data['app_hostname']
        create_run_linux = data['create_run_linux']
        mongo_version = float(data['software_version'])
        # repository = network or local
        repository = data['repository']
        org_id = node['org_id']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,release['stdout'],release['stderr'],release['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,release['stdout'],release['stderr'],release['excode'])
            update_failed_object(node['id'])
            return
        
        version_exists = check_mongo_version_exists(host, mongo_version, distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not version_exists:
        step = 'installing mongodb'
        duration_step_start = time.perf_counter()

        if Repository.LOCAL.value in repository:
            jsn = install_offline_remote_to_remote(distributor, release, host, mongo_version, DBType.MONGO.value, app_host)
        else:
            jsn = install_mongo_online(distributor, release, host, mongo_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating node'
    duration_step_start = time.perf_counter()

    jsn = setup_node_directory(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    jsn = start_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'configuring authentication'
    duration_step_start = time.perf_counter()

    error = create_superuser(host, node, shard = True)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    data_path = node['dir_123Cluster'].strip()
    if data_path.startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        tilda_path = jsn['stdout']
        data_path = data_path.replace('~', tilda_path)

    jsn = stop_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    if use_ssl:
        shard_primary_ssl_root = f"{data_path}/data"
        jsn = transfer_ssl_certificates(config_node, config_host, host, shard_primary_ssl_root)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        jsn = inject_ssl_config(host, data_path)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        jsn = write_ssl_to_file(host, shard_primary_ssl_root, node['port'])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
     
    # authorization must be added only after the create superuser and configure ssl (and also to avoid editing config of running node)
    try:
        config_content = f"""
security:
    authorization: enabled
    keyFile: {data_path}/conf/key.file
"""
        if mongo_version >= 7.0:
            config_content += """
    clusterAuthMode: keyFile
"""
        source_keyfile = f"{config_node['dir_123Cluster'].strip()}/conf/key.file"
        target_keyfile = f'{data_path}/conf/key.file'
        if host['name'].strip() == config_host['name'].strip():
            try:
                jsn = json.loads(remote_command(config_host, f"sudo rsync -a '{source_keyfile}' '{target_keyfile}'", True), strict = False)
                if jsn['stderr'] != '' or jsn['stdout'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return
        else:
            jsn = json.loads(remote_command(config_host, f"sudo cat '{source_keyfile}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            keyfile_content = jsn['stdout']
            jsn = json.loads(remote_command(host, f"echo '{keyfile_content}' | sudo tee '{target_keyfile}' && sudo chmod 400 '{target_keyfile}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        remote_command(host, f"echo '{config_content}' | sudo tee -a '{data_path}/conf/mongod.cfg'", True)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    jsn = start_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'initializing replicaset'
    duration_step_start = time.perf_counter()

    add_to_cfg(host, node, shard_rs=True)

    jsn = restart_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    error = add_first_node_to_replicaset(host, node)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    

    step = 'joining shard to cluster'
    duration_step_start = time.perf_counter()
   
    try:
        
        client = create_mongo_connection(config_node, config_host['name'], mongos_port=mongos_port, direct=True)
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            return
        
        with client:
                client.admin.command("addShard", f"{node['ex_params'].get('replicaset_name')}/{host['name']}:{node['port']}")
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    
    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024

        remote_command(host, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task', 'monitor_name': None,  'org_id': node['org_id'],  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, cluster_role=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['version'], ClusterRoles.PRIMARY.value, node_size, node['id'],))
                if node['ex_params'].get('config_replicaset_name'):
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"severity":\"%s\"}' WHERE id=%s""" % (Severity.OK.value, node['cluster_id'],))
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"status":\"\"}' WHERE id=%s""", (node['cluster_id'],))
                
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()

                
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)

    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])

    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as err:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def start_mongo_node_task(data, task_id):
    """start a Mongo Node"""

    task = 'start mongo node task'

    try:
        # data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        secondary_nodes = data.get('secondary_data')
        has_mongos = data['has_mongos']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting node'
    duration_step_start = time.perf_counter()

    jsn = restart_node(host, node, secondary_nodes, move_databases = not has_mongos)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)


def install_mongo_online(distributor, release, host, mongo_version) -> dict:
    jsn = {'stdout': '', 'stderr': '', 'excode': 0}
    try:
        supported_versions = SupportedOS[distributor.upper()].value
        supported_versions_str = ", ".join(map(str, supported_versions))
        if OperatingSystem.CENTOS.value in distributor.lower() or OperatingSystem.REDHAT.name.lower() in distributor.lower():
            package_manager = 'dnf' if int(release) > 7 else 'yum'
            if int(release) in supported_versions:
                mongo_repo = f"""[mongodb-org-{mongo_version}]
name=MongoDB Repository
baseurl=https://repo.mongodb.org/yum/{OperatingSystem.REDHAT.name.lower()}/$releasever/mongodb-org/{mongo_version}/x86_64/
gpgcheck=1
enabled=1
gpgkey=https://www.mongodb.org/static/pgp/server-{mongo_version}.asc"""

                jsn = json.loads(remote_command(host, f"echo '{mongo_repo}' | sudo tee '/etc/yum.repos.d/mongodb-org.repo'", True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'] if 'nothing to do' not in error_lower))
                if filter:
                    return jsn

                jsn = json.loads(remote_command(host, f'sudo {package_manager} install -q -y mongodb-org rsync tar net-tools', True), strict=False)
                return jsn

        elif OperatingSystem.UBUNTU.value in distributor.lower():
            commands = [
                "sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y gpg net-tools",
                # apt-key are deprecated since ubuntu 20.11 - always use gpg to install repository keys
                f"wget -qO - https://www.mongodb.org/static/pgp/server-{mongo_version}.asc | sudo gpg -q --batch --yes --dearmor -o '/usr/share/keyrings/mongodb-{mongo_version}-keyring.gpg'"
            ]

            if int(release) == 18 or mongo_version == 4.2:
                commands.append(f"echo 'deb [arch=amd64,arm64 signed-by=/usr/share/keyrings/mongodb-{mongo_version}-keyring.gpg] https://repo.mongodb.org/apt/ubuntu bionic/mongodb-org/{mongo_version} multiverse' | sudo tee '/etc/apt/sources.list.d/mongodb-org.list'")
            elif int(release) == 20:
                commands.append(f"echo 'deb [arch=amd64,arm64 signed-by=/usr/share/keyrings/mongodb-{mongo_version}-keyring.gpg] https://repo.mongodb.org/apt/ubuntu focal/mongodb-org/{mongo_version} multiverse' | sudo tee '/etc/apt/sources.list.d/mongodb-org.list'")
            elif int(release) == 22:
                commands.append(f"echo 'deb [arch=amd64,arm64 signed-by=/usr/share/keyrings/mongodb-{mongo_version}-keyring.gpg] https://repo.mongodb.org/apt/ubuntu jammy/mongodb-org/{mongo_version} multiverse' | sudo tee '/etc/apt/sources.list.d/mongodb-org.list'")
            else:
                err_msg = f'Release version of {distributor.capitalize()} is not {supported_versions_str}'
                jsn = {}
                jsn['stdout'] = None
                jsn['stderr'] = err_msg
                jsn['excode'] = 1
                return jsn

            commands += ['sudo apt-get update', 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y mongodb-org']

            for command in commands:
                jsn = json.loads(remote_command(host, command, True), strict=False)
                if jsn['stdout'] != '':
                    print('stdout:', jsn['stdout'])
                if jsn['stderr'] != '':
                    print('stderr:', jsn['stderr'])

                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                if filter:
                    return jsn
        else:
            err_msg = 'Distributor is not Ubuntu or CentOS'
            jsn = {}
            jsn['stdout'] = None
            jsn['stderr'] = err_msg
            jsn['excode'] = 1
            return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

    return jsn

def setup_node_directory(host, node, datadir_only=False) -> dict:
    """create data, log and config directories of instance with needed permissions
       datadir_only - prepare only data subfolder (for rebuild the structure after delete_node_data call)
    """

    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        data_path = node['dir_123Cluster']

        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        create_folder_commands = [
            f"sudo mkdir -p '{data_path}/data'",
        ]

        if not datadir_only:
            create_folder_commands += [
                f"sudo mkdir -p '{data_path}/log'",
                f"sudo mkdir -p '{data_path}/conf'",
            ]

        for command in create_folder_commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                return jsn

        if not datadir_only:
            jsn = set_mongo_config_file(host, node)

            if jsn['stderr'] != '':
                return jsn

        jsn = set_mongo_permissions(host, node, datadir_only)
        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def set_mongo_permissions(host, node, after_wipe=False) -> dict:
    """set mongo recommended permissions on node directory
        after_wipe - if we reset permissions after data directory wipe
    """

    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        data_path = node['dir_123Cluster'].strip()

        set_permissions_commands = [
            f"sudo chown -R $(whoami) '{data_path}'",
            f"sudo chmod -R 755 '{data_path}'"
        ]
        if after_wipe:
            set_permissions_commands.append(f"sudo chmod 400 '{data_path}/conf/key.file'")
        
        for command in set_permissions_commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                return jsn

        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def set_mongo_config_file(host, node, mongos=False) -> dict:
    """create mongod.cfg file"""

    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        data_path = node['dir_123Cluster']
        cluster_role = node['cluster_role']
        unit = "mongos" if mongos else 'mongod'
        mongo_version = float(node['version'])
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        conf_file = f"{data_path}/conf/{unit}.cfg"
        # left align - easier to match indents
        config_content = f"""\
systemLog:
    destination: file
    path: {data_path}/log/{unit}.log
    logAppend: true
    logRotate: reopen
net:
    bindIp: {host['name']}, 127.0.0.1
processManagement:
    fork: true
"""
        # arbiters and mongos are without data
        if not any(r in cluster_role for r in [ClusterRoles.ARBITER.value]) and not mongos:
            config_content += f"""\
storage:
    dbPath: {data_path}/data
"""
            # for primary, initial config should exclude security -
        # should be configured separately after superuser creation
        if any(r in cluster_role for r in [ClusterRoles.ARBITER.value, ClusterRoles.STANDBY.value]) and not mongos:
            config_content += f"""\
security:
    {'authorization: enabled'  if not mongos else ''}
    keyFile: {data_path}/conf/key.file
"""
            if mongo_version >= 7.0:
                config_content += """
    clusterAuthMode: keyFile
"""        
        
        if mongos:
            replicaset_name = node['ex_params'].get('replicaset_name')
            config_content += f"""
sharding:
    configDB: {replicaset_name}/{host['name']}:{node['port']}
"""
        # do not append, rewrite
        jsn = json.loads(remote_command(host, f"echo '{config_content}' | sudo tee '{conf_file}'", True), strict=False)

        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def start_node(host, node, secondary_nodes=None, in_replicaset=False) -> dict:
    """start a Mongo Node (default - as standalone)"""

    try:
        replicaset_name = node['ex_params'].get('replicaset_name')
        mongo_version = float(node['version'])
        data_path = node['dir_123Cluster']
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        mongo_bin = find_mongo_bin(host, mongo_version)
        command = f"""sudo {mongo_bin}/mongod --fork --config '{data_path}/conf/mongod.cfg' --port '{node['port']}' --dbpath '{data_path}/data' --bind_ip '{host['name']}' {f"--replSet '{replicaset_name}'" if in_replicaset else ''}"""
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

    if secondary_nodes:
        (is_error, result) = find_replicaset_primary(host, node)
        if is_error:
            jsn['stderr'] = result
            return jsn

        error = update_cluster_roles(node, result, secondary_nodes)
        if error != '':
            jsn['stderr'] = error
            return jsn

    return jsn

def create_superuser(host, node, shard=False, mongos = False) -> str:
    """create 123Cluster app superuser for primary node with no security configured"""

    try:
        roles = ['userAdminAnyDatabase', 'root', 'readWriteAnyDatabase', 'super_role']
        if shard and float(node['version']) >= 8.0:
            roles.append('directShardOperations')
        # there could be races in configuring node, so update or create
        client = create_mongo_connection(node, host['name'], ReadPreference.PRIMARY_PREFERRED.value, no_auth=True, no_ssl=True, direct=True, mongos_port=node['ex_params'].get('mongos_port') if mongos else None)
        if isinstance(client, tuple):
            return(str(client[1]))

        with client:
            role_exists = any(r for r in client.admin.command('rolesInfo')['roles'] if r['role'] == 'super_role')
            if role_exists:
                client.admin.command('updateRole', 'super_role', privileges=[{"resource" : {"anyResource" : True}, "actions" : ["anyAction"]}], roles=[])
            else:
                client.admin.command('createRole', 'super_role', privileges=[{"resource" : {"anyResource" : True}, "actions" : ["anyAction"]}], roles=[])

            user_exists = any(u for u in client.admin.command('usersInfo')['users'] if u['user'].strip() == node['database_user'].strip())
            if user_exists:
                client.admin.command('updateUser', node['database_user'], pwd=node['database_pwd'], roles = roles)
            else:
                client.admin.command('createUser', node['database_user'], pwd=node['database_pwd'], roles = roles)

            return ''
    except Exception as e:
        error, = e.args
        return str(error)

def delete_mongo_standalone_node_task(data, task_id):
    """task to delete a standalone node"""

    task = 'delete standalone node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        # for situation: have just primary and arbiter so need to drop arbiter first
        has_arbiter = data['has_arbiter']
        if has_arbiter:
            # arbiter_data=[{host:..., arbiter:...}, {host:..., arbiter:...}, ...]
            arbiter_data = data['arbiter_data']
            arbiters_id = data['arbiters_id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node existence'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if jsn['stdout'] == '':
            # if port does not occupied - can remove entire node folder
            jsn = json.loads(remote_command(host, f"sudo rm -rf '{node['dir_123Cluster']}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is already deleted',1)

            try:
                with get_postgres_connection() as  conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node["id"],))
                        conn.commit()
            except Exception as error:
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            step = 'task finished'
            status = Severity.WARNING.value
            duration = 0
            logger(task_id,task_log_id,task,step,status,duration,created_by)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if has_arbiter:
        step = 'removing arbiter'
        duration_step_start = time.perf_counter()

        for arb in arbiter_data:
            jsn = stop_node(arb['host'], json.loads(arb['arbiter']))
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            jsn = delete_node_data(arb['host'], json.loads(arb['arbiter']))
            if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in ['no such', 'no found']):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'removing node'
    duration_step_start = time.perf_counter()

    jsn = stop_node(host, node)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    # if we drop living node - keep its config + logs (subject of change?)
    jsn = delete_node_data(host, node)
    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in ['no such', 'no found']):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id']))
                if has_arbiter:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=ANY(%s)""", (Severity.NOT_OK.value, Status.DELETED.value, arbiters_id))

                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false',  ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node["id"]))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_node(host, node, secondary_nodes = None) -> dict:
    """function to stop mongo node daemon"""

    mongo_version = float(node['version'])
    data_path = node['dir_123Cluster']
    if data_path.startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            return jsn

        tilda_path = jsn['stdout']
        data_path = data_path.replace('~', tilda_path)

    mongo_bin = find_mongo_bin(host, mongo_version)
    try:
        command = f"sudo {mongo_bin}/mongod --config '{data_path}/conf/mongod.cfg' --port '{node['port']}' --dbpath '{data_path}/data' --bind_ip '{host['name']}' --shutdown"
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in ["server running with dbpath", "does not appear to be running", "invalid process id"]):
            jsn['stderr'] = ''
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

    # if we have orphaned secondary nodes - try to find who is new primary
    if secondary_nodes:
        secondary_nodes = [{'host': item['host'], 'node': json.loads(item['node'])} for item in secondary_nodes]
        error = ''
        primary = ''
        for secondary in secondary_nodes:
            try:
                (is_error, result) = find_replicaset_primary(secondary['host'], secondary['node'])
                if is_error:
                    error = result
                else:
                    primary = result
                    break
            except Exception as e:
                error = str(e.args[0])
        else:
            if error:
                jsn['stderr'] = error
                return jsn

        if primary:
            error = update_cluster_roles(node, result, secondary_nodes)
            if error != '':
                jsn['stderr'] = error
                return jsn

    return jsn

def delete_node_data(host, node) -> dict:
    """wipe the data directory of node (config and logs stay intact)"""

    data_path = node['dir_123Cluster']
    if data_path.startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            return jsn

        tilda_path = jsn['stdout']
        data_path = data_path.replace('~', tilda_path)

    try:
        jsn = json.loads(remote_command(host, f"sudo rm -rf '{data_path}/data'", True), strict=False)
    except Exception as error:
        # don't kill the task because if fail to remove - task will be killed on later steps anyway
        return {'stdout': None, 'stderr': '', 'excode': 9}

    return jsn

def delete_mongo_replicaset_node_task(data, task_id):
    """task to delete a replicaset node"""

    task = 'delete replicaset node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        final_backup = data.get('final_backup')
        backup_host = data['backup_host']
        secondary_nodes = data.get('secondaries_data')
        arbiters =  data.get('arbiters_data')
        arbiters_id = data.get('arbiters_id')
        standbys = data.get('standbys')
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        has_arbiter = isinstance(arbiters_id, list) and len(arbiters_id) > 0
        data_path = node['dir_123Cluster']
        mongos_node = data.get('mongos_node')
        mongos_host = data.get('mongos_host')
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in standbys]

        if has_arbiter:
            arbiters = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in arbiters]

        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        connect_host = None
        connect_node = None
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node existence'
    duration_step_start = time.perf_counter()

    try:
        is_started = False
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        # if port are not occupied - can just remove whole directory of node
        if jsn['stdout'] == '':
            jsn = json.loads(remote_command(host, f"sudo rm -rf '{node['dir_123Cluster']}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is already deleted', 1)

            try:
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node["id"],))
                        conn.commit()
            except Exception as error:
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            step = 'task finished'
            status = Severity.WARNING.value
            duration = 0
            logger(task_id,task_log_id,task,step,status,duration,created_by)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
            return
        else:
            is_started = True
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    #At that point, node may be up but broken -
    #we can drop node at such condition, so dont stop on find primary error
    started_but_down = False
    is_error, result = find_replicaset_primary(host, node)
    if is_error and is_started:
        # means, started, but down
        started_but_down = True

    is_primary = result == node['name']

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # if node is primary and has standbys - try to stepdown (break with red step on error, since unsuccessful stepdown may be breaking action)
    if is_primary and has_standby:
        step = 'stepping down primary'
        duration_step_start = time.perf_counter()

        # here must stop if fail,
        # if mongo does not allow to stepdown - should be a reason
        error = step_down_node(host, node)
        if error != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return

        # give enough time for primary re-election complete
        time.sleep(10)
        (is_error, result) = find_replicaset_primary(host, node)
        if is_error:
            for i in standbys:
                (is_error, result) = find_replicaset_primary(i['host'], i['node'])
                if not is_error:
                    break

        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
            return

        error = update_cluster_roles(node, result, secondary_nodes, move_databases= not mongos_node)
        if error != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        new_primary = next((i for i in standbys if i['node']['name'] in result), None)
        if new_primary:
            connect_host = new_primary['host']
            connect_node = new_primary['node']

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'searching for primary'
    duration_step_start = time.perf_counter()

    try:
        #if not found primary at that point - use failback values from task to start discovery
        if not connect_host:
            connect_host = data.get('host_primary')
            if connect_host:
                connect_host = json.loads(connect_host)

        if not connect_node:
            connect_node = data.get('primary_data')
            if connect_node:
                connect_node = json.loads(connect_node)

        # find primary within available nodes
        if has_standby:
            for i in standbys:
                (is_error, result) = find_replicaset_primary(i['host'], i['node'])
                if not is_error:
                    break

            primary = next((i for i in standbys if i['node']['name'].strip() == result), None)
            if primary:
                connect_host = primary['host']
                connect_node = primary['node']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # if node is primary and has arbiters but dont have standbys - try to drop all arbiters
    # (continue through error since we removing primary anyway)
    if is_primary and not has_standby and has_arbiter:
        step = 'dropping arbiters'
        duration_step_start = time.perf_counter()

        try:
            last_known_error = ''
            if connect_host and connect_node:
                for arbiter in arbiters:
                    arbiter_host = arbiter['host']
                    arbiter_node = arbiter['node']
                    error = remove_node_from_rs_using_primary(connect_host, connect_node, arbiter_host['name'], arbiter_node['port'])
                    if error != '':
                        last_known_error = f"{last_known_error}{error}\n"
                    error = stop_node(arbiter_host, arbiter_node)
                    if error['stderr'] != '':
                        last_known_error = f"{last_known_error}{error['stderr']}\n"
                    error = delete_node_data(arbiter_host, arbiter_node)
                    if error['stderr'] != '':
                        last_known_error = f"{last_known_error}{error['stderr']}\n"
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

        status = Severity.OK.value if not last_known_error else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if not last_known_error else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,last_known_error,1)
    
    # before running rs.remove() on primary for target node - need to stop target node first
    # https://www.mongodb.com/docs/v7.0/tutorial/remove-replica-set-member/
    # but this will cause re-election, so do opposite

    # if started but down, rs.remove() go stuck and rs.reconfig() just give false positive - so skip
    if not started_but_down:
        step = 'removing node from replicaset'
        duration_step_start = time.perf_counter()

        if connect_host and connect_node:
            error = remove_node_from_rs_using_primary(connect_host, connect_node, host['name'], node['port'])
            if error != '':
                # do not stop here, removing node from replica set just recommended good practice
                status = Severity.WARNING.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error, 1)
            else:
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        # if cant found primary - proceed with drop anyway, but generate warning
        else:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'no primary found',1)

        step = 'stopping node'
        duration_step_start = time.perf_counter()

        jsn = stop_node(host, node) if not started_but_down else stop_mongo_process(host, node['port'])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        if ClusterRoles.BALANCER.value in node['cluster_role']:
            jsn = stop_mongo_process(host, node['ex_params'].get('mongos_port'))
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
    if node['cluster_role'] != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        try:
            step = 'backing up the node'
            duration_step_start = time.perf_counter()

            backup_path =  final_backup['ex_params'].get('backup_path')

            jsn = json.loads(remote_command(host, f"sudo mkdir -p '{backup_path}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            cmd = f"sudo tar zcfC '{backup_path}/{final_backup['name']}.tar.gz' '{node['dir_123Cluster']}' data"
            jsn = json.loads(remote_command(host, cmd, True), strict=False)

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and 'removing leading' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            jsn = transfer_files_rsync(host, backup_host, backup_path, [f"{backup_path}/"])
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            jsn = json.loads(local_command(f"""sudo du -b '{backup_path}{f'/{final_backup["name"]}.tar.gz'}' | awk '{{print $1}}'""", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            size = jsn['stdout'].strip()
            backup_size = float(size)/1024

            modified_date = get_date_modified_folder(folder_path=backup_path)
            if isinstance(modified_date, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
                return

            remote_command(host, f"sudo rm -rf '{backup_path}'", True)

            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,final_backup['id']))
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, final_backup['id']))
                    conn.commit()

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    step = 'removing node data'
    duration_step_start = time.perf_counter()

    # if we drop living node - keep its config + logs (subject of change?)
    jsn = delete_node_data(host, node)
    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in ['no such', 'no found']):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if ClusterRoles.BALANCER.value in node['cluster_role']:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_id=NULL, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                    if not has_standby and not has_arbiter:
                        cur.execute("""DELETE FROM "clusterCore_c123_cluster" WHERE id=%s""", (node['cluster_id']))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false',  ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node["id"],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def find_replicaset_primary(host, node, max_retries = 1) -> tuple:
    """ find name of current replica set primary"""

    try:
        client = create_mongo_connection(node, host['name'], ReadPreference.PRIMARY_PREFERRED.value)
        if isinstance(client, tuple):
            return (True, str(client[1]))

        with client:
            # check if node client are actually up
            for _ in range(max_retries):
                try:
                    ping = client.admin.command('ping').get('ok', 0.0)
                    if float(ping) == 1.0:
                        break
                except Exception as e:
                    time.sleep(1)
                    error = str(e.args[0])
            else:
                if error != '':
                    return (True, error)

            # get base server info
            info = client.admin.command('hello')
            rs = info.get('setName')
            # if node is not secondary/arbiter, but not in replica set - it is standalone
            if client.is_primary and not rs:
                return (False, 'standalone')
            elif rs:
                # if node are part of replica set - get name of replica set primary
                primary = info.get('primary')
                if primary:
                    return (False, primary.strip())

            # if got here - replica set dont have primary
            return (True, 'not able to find primary')
    except Exception as e:
        error, = e.args
        return (True, str(error))

def step_down_node(host, node) -> str:
    """step down a primary node in a replica set"""

    try:
        client = create_mongo_connection(node, host['name'])
        if isinstance(client, tuple):
            return str(client[1])

        with client:
            # 1. Dont use force on mongo commands : good practice is avoid transaction rollback
            # 2. Instead of force, enlarge secondary catch up period
            # (but it should be lower than overall stepdown time)
            # 3. In complex cases of severely lagging replica set,
            # and if we know who of members has most fresh oplog - perform manual stepdown,
            # with prior running replSetFreeze on other electable secondaries
            client.admin.command({'replSetStepDown': 120, 'secondaryCatchUpPeriodSecs': 80})
            return ''
    except Exception as e:
        error, = e.args
        return str(error)

def physical_backup_mongo_node_task(data, task_id):
    """a task that backs-up a node to a path"""

    task = 'physical backup mongo node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        backup_host = data['backup_host']
        node = json.loads(data['node_data'])
        backup = json.loads(data['backup_data'])
        compress = backup["ex_params"].get("compress")
        data_path = node['dir_123Cluster']
        schedule = backup['is_sched']
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        backup_path = backup['ex_params'].get('backup_path')
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)

        if schedule:
            oplog_path = f"{backup_path}/dumps_oplog"
            start_date = backup['name'].split('_')[1]
            backup_path = f"{backup_path}/{start_date}"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'backing up the node'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo mkdir -p '{backup_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    try:
        client = create_mongo_connection(node, host['name'])
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            return

        with client:
            locked = float(client.admin.command("fsync", lock=True).get("ok", 0.0))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    if locked != 1.0:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,"could not lock node against writes",1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,"could not lock node against writes",1)
        return

    try:
        # Go to directory of node and add only data sub-folder
        commands = [
            f"sudo rsync -a --recursive --exclude=diagnostic.data --exclude=*.lock  '{data_path}/data' '{backup_path}'"
        ]
        if compress:
            commands.insert(1, f"""sudo tar zcfC '{backup_path}/{backup['name']}.tar.gz' '{backup_path}' data""")
            commands.insert(2, f"sudo rm -rf '{backup_path}/data'")

        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if (jsn['stderr'] != '' and 'removing leading' not in jsn['stderr'].lower()) or ('rsync' in command and jsn['stdout'] != ''):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    lockCount = -1
    try:
        client = create_mongo_connection(node, host['name'])
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            return

        with client:
            while lockCount != 0:
                lockCount = int(client.admin.command("fsyncUnlock").get('lockCount', -1))
                if lockCount < 0:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,"could not unlock node for writes",1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,"could not unlock node for writes",1)
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,"could not unlock node for writes",1)
                    return
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying backup of node to backup host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(host, backup_host, backup_path, [f"{backup_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if schedule:
            jsn = check_local_directory_exists(oplog_path, 'd')
            if jsn['stderr'] != '':
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            exists = jsn['stdout']
            if 'false' in exists:
                try:
                    #to maintain expected backup structure
                    jsn = json.loads(local_command(f"sudo mkdir -p '{oplog_path}'", True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                except Exception as error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                    return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking the backup size'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(local_command(f"sudo du -sb '{backup_path}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        

        size = jsn['stdout'].strip()
        backup_size = float(size)/1024

        modified_date = get_date_modified_folder(folder_path=backup['ex_params'].get('backup_path'))
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return

        jsn = json.loads(remote_command(host, f"sudo rm -rf '{backup_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, backup['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end,backup['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_mongo_node_task(data, task_id):
    """task to stop mongo node deamon"""

    task = 'stop mongo node task'

    try:
        # data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        secondary_nodes = data.get('secondary_data')
        data_path = node['dir_123Cluster']
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping node'
    duration_step_start = time.perf_counter()

    jsn = stop_node(host, node, secondary_nodes)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=%s""", (Severity.NOT_OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)




def restart_mongo_node_task(data, task_id):
    """restart a Mongo Node"""

    task = 'restart mongo node task'

    try:
        # data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        secondary_nodes = data.get('secondary_data')
        has_mongos = data['has_mongos']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restarting node'
    duration_step_start = time.perf_counter()

    jsn = restart_node(host, node, secondary_nodes, move_databases= not has_mongos)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true'WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def upgrade_mongo_node_task(data, task_id):
    """upgrade a Mongo Node to newer major version"""

    task = 'upgrade mongo node task'

    try:
        # data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        target_version = data['target_version']
        app_host = json.loads(data['app_host'])
        nodes = [{'host': json.loads(n['host']), 'node': json.loads(n['node'])} for n in data['nodes']]
        arbiters = [n for n in nodes if ClusterRoles.ARBITER.value in n['node']['cluster_role']]
        for n in nodes:
            if ClusterRoles.ARBITER.value in n['node']['cluster_role']:
                continue
            
            (is_error, result) = find_replicaset_primary(n['host'], n['node'])
            if not is_error:
                primary_name = result
                break
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find a primary',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find a primary',1)
            return
        
        primary = next((n for n in nodes if n['node']['name'].strip() == primary_name.strip()), None)
        if not primary:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find a primary',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find a primary',1)
            return
        
        other_than_primary = [n for n in nodes if n['node']['name'].strip() != primary['node']['name'].strip()]
        
        repository = primary['node']['ex_params']['repository']
        primary_version = primary['node']['version']
        use_ssl = primary['node']['ex_params'].get('ssl')
        distributor, release = check_os_version(primary['host'])
        if OperatingSystem.INVALID.value in distributor:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,release['stdout'],release['stderr'],release['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,release['stdout'],release['stderr'],release['excode'])
            return
    except Exception as error:
        print(traceback.format_exc())
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'setting primary feature compatibility version'
    duration_step_start = time.perf_counter()

    try:
        client = create_mongo_connection(primary['node'], primary['host']['name'], ReadPreference.PRIMARY.value)
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(client[1]),1)
            return
        
        set_primary_compatibility_version = {"setFeatureCompatibilityVersion": str(primary_version)}
        if float(primary_version) > 6.0:
              set_primary_compatibility_version['confirm'] = True
        
        client.admin.command(set_primary_compatibility_version)
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    step = 'upgrading mongodb binaries'
    duration_step_start = time.perf_counter()

    last_known_err = ''
    
    try:
        for n in nodes:
            version_exists = check_mongo_version_exists(n['host'], target_version, distributor)
            if not version_exists:
                if OperatingSystem.CENTOS.value in distributor:
                    check_mongo_exists_cmd = "sudo rpm -qa | grep 'mongo'"
                else:
                    check_mongo_exists_cmd = "sudo dpkg -l | grep 'mongo'"
                jsn = json.loads(remote_command(n['host'], check_mongo_exists_cmd, True), strict=False)
                mongo_exists = jsn['stdout'] != ''
                if mongo_exists:
                    if OperatingSystem.CENTOS.value in distributor:
                        package_manager = 'dnf' if int(release) > 7 else 'yum'
                        remove_mongo_cmd = f"sudo {package_manager} -q -y remove mongodb-org"
                    else:
                        remove_mongo_cmd = "sudo apt-get -qq -y purge ^mongo"
                    
                    jsn = json.loads(remote_command(n['host'], remove_mongo_cmd, True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                    if filter:
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    
                if Repository.LOCAL.value in repository:
                    jsn = install_offline_remote_to_remote(distributor, release, n['host'], target_version, DBType.MONGO.value, app_host)
                else:
                    jsn = install_mongo_online(distributor, release, n['host'], target_version)

                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                if filter:
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    except Exception as error:
        logging.error(f"{task}@{step}: {task_id}: {str(error)}")
                
    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    step = 'restarting non-primary nodes'
    duration_step_start = time.perf_counter()

    last_known_err = ''

    try:
        for n in other_than_primary:
            jsn = restart_node(n['host'], n['node'])
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")    
    except Exception as error:
        logging.error(f"{task}@{step}: {task_id}: {str(error)}")
    
    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'stepdown primary'
    duration_step_start = time.perf_counter()

    try:
        error = step_down_node(primary['host'], primary['node'])
        if error != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'searching for new primary'
    duration_step_start = time.perf_counter()

    (is_error, primary_name) = find_replicaset_primary(primary['host'], primary['node'])
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,primary_name,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,primary_name,1)
        return
        
    error = update_cluster_roles(primary['node'], result, [n['node'] for n in nodes if n['node']['name'].strip() != primary_name.strip()])
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restarting the old primary'
    duration_step_start = time.perf_counter()

    try:
        jsn = restart_node(primary['host'], primary['node'])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'setting target feature compatibility version'
    duration_step_start = time.perf_counter()

    try:
        primary = next((n for n in nodes if n['node']['name'].strip() == primary_name.strip()), None)
        if not primary:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find a primary',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find a primary',1)
            return

        client = create_mongo_connection(primary['node'], primary['host']['name'], ReadPreference.PRIMARY.value)
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(client[1]),1)
            return
        
        set_target_compatibility_version = { "setFeatureCompatibilityVersion": str(target_version)}
        if float(target_version) > 6.0:
            set_target_compatibility_version['confirm'] = True
        
        client.admin.command(set_target_compatibility_version)
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    if float(primary_version) < 4.4 and len(arbiters) > 0:
        step = 'rejoin arbiter nodes'
        duration_step_start = time.perf_counter()

        last_known_err = ''

        try:
            for a in arbiters:
                jsn = stop_node(a['host'], a['node'])
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                if use_ssl:
                    jsn = transfer_ssl_certificates(a['node'], a['host'], a['host'], mode='backup')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                jsn = delete_node_data(a['host'], a['node'])
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                jsn = setup_node_directory(a['host'], a['node'], datadir_only = True)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                if use_ssl:
                    jsn = transfer_ssl_certificates(a['node'], a['host'], a['host'], mode='restore')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                jsn = start_node(a['host'], a['node'], in_replicaset=False)
                if jsn['stderr'] != '' and f"server running with dbpath:" not in jsn['stderr'].lower():
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                jsn = restart_node(a['host'], a['node'])
                if jsn['stderr'] != '' and f"server running with dbpath:" not in jsn['stderr'].lower():
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                error = using_primary_add_node_to_replicaset(primary['host'], primary['node'], a['host'], a['node'])
                if error != '':
                    last_known_err = error
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")

        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    try:
       with get_postgres_connection() as conn:
           with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true' WHERE cluster_id=%s """, (Severity.OK.value, Status.OK.value, target_version, primary['cluster_id']))
                conn.commit()
    except Exception as error:
       step = 'null'
       status = Severity.NOT_OK.value
       duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
       logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
       return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def upgrade_mongo_node_task(data, task_id):
    """upgrade a Mongo Node to newer major version"""

    task = 'upgrade mongo node task'

    try:
        # data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        target_version = data['target_version']
        app_host = json.loads(data['app_host'])
        nodes = [{'host': json.loads(n['host']), 'node': json.loads(n['node'])} for n in data['nodes']]
        arbiters = [n for n in nodes if ClusterRoles.ARBITER.value in n['node']['cluster_role']]
        for n in nodes:
            if ClusterRoles.ARBITER.value in n['node']['cluster_role']:
                continue
            
            (is_error, result) = find_replicaset_primary(n['host'], n['node'])
            if not is_error:
                primary_name = result
                break
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find a primary',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find a primary',1)
            return
        
        primary = next((n for n in nodes if n['node']['name'].strip() == primary_name.strip()), None)
        if not primary:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find a primary',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find a primary',1)
            return
        
        other_than_primary = [n for n in nodes if n['node']['name'].strip() != primary['node']['name'].strip()]
        
        repository = primary['node']['ex_params']['repository']
        primary_version = primary['node']['version']
        use_ssl = primary['node']['ex_params'].get('ssl')
        distributor, release = check_os_version(primary['host'])
        if OperatingSystem.INVALID.value in distributor:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,release['stdout'],release['stderr'],release['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,release['stdout'],release['stderr'],release['excode'])
            return
    except Exception as error:
        print(traceback.format_exc())
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'setting primary feature compatibility version'
    duration_step_start = time.perf_counter()

    try:
        client = create_mongo_connection(primary['node'], primary['host']['name'], ReadPreference.PRIMARY.value)
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(client[1]),1)
            return
        
        set_primary_compatibility_version = {"setFeatureCompatibilityVersion": str(primary_version)}
        if float(primary_version) > 6.0:
              set_primary_compatibility_version['confirm'] = True
        
        client.admin.command(set_primary_compatibility_version)
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    step = 'upgrading mongodb binaries'
    duration_step_start = time.perf_counter()

    last_known_err = ''
    
    try:
        for n in nodes:
            version_exists = check_mongo_version_exists(n['host'], target_version, distributor)
            if not version_exists:
                if OperatingSystem.CENTOS.value in distributor:
                    check_mongo_exists_cmd = "sudo rpm -qa | grep 'mongo'"
                else:
                    check_mongo_exists_cmd = "sudo dpkg -l | grep 'mongo'"
                jsn = json.loads(remote_command(n['host'], check_mongo_exists_cmd, True), strict=False)
                mongo_exists = jsn['stdout'] != ''
                if mongo_exists:
                    if OperatingSystem.CENTOS.value in distributor:
                        package_manager = 'dnf' if int(release) > 7 else 'yum'
                        remove_mongo_cmd = f"sudo {package_manager} -q -y remove mongodb-org"
                    else:
                        remove_mongo_cmd = "sudo apt-get -qq -y purge ^mongo"
                    
                    jsn = json.loads(remote_command(n['host'], remove_mongo_cmd, True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                    if filter:
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    
                if Repository.LOCAL.value in repository:
                    jsn = install_offline_remote_to_remote(distributor, release, n['host'], target_version, DBType.MONGO.value, app_host)
                else:
                    jsn = install_mongo_online(distributor, release, n['host'], target_version)

                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                if filter:
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    except Exception as error:
        logging.error(f"{task}@{step}: {task_id}: {str(error)}")
                
    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    step = 'restarting non-primary nodes'
    duration_step_start = time.perf_counter()

    last_known_err = ''

    try:
        for n in other_than_primary:
            jsn = restart_node(n['host'], n['node'])
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")    
    except Exception as error:
        logging.error(f"{task}@{step}: {task_id}: {str(error)}")
    
    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'stepdown primary'
    duration_step_start = time.perf_counter()

    try:
        error = step_down_node(primary['host'], primary['node'])
        if error != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'searching for new primary'
    duration_step_start = time.perf_counter()

    (is_error, primary_name) = find_replicaset_primary(primary['host'], primary['node'])
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,primary_name,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,primary_name,1)
        return
        
    error = update_cluster_roles(primary['node'], result, [n['node'] for n in nodes if n['node']['name'].strip() != primary_name.strip()], move_databases = True)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restarting the old primary'
    duration_step_start = time.perf_counter()

    try:
        jsn = restart_node(primary['host'], primary['node'])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'setting target feature compatibility version'
    duration_step_start = time.perf_counter()

    try:
        primary = next((n for n in nodes if n['node']['name'].strip() == primary_name.strip()), None)
        if not primary:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find a primary',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find a primary',1)
            return

        client = create_mongo_connection(primary['node'], primary['host']['name'], ReadPreference.PRIMARY.value)
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(client[1]),1)
            return
        
        set_target_compatibility_version = { "setFeatureCompatibilityVersion": str(target_version)}
        if float(target_version) > 6.0:
            set_target_compatibility_version['confirm'] = True
        
        client.admin.command(set_target_compatibility_version)
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    if float(primary_version) < 4.4 and len(arbiters) > 0:
        step = 'rejoin arbiter nodes'
        duration_step_start = time.perf_counter()

        last_known_err = ''

        try:
            for a in arbiters:
                jsn = stop_node(a['host'], a['node'])
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                if use_ssl:
                    jsn = transfer_ssl_certificates(a['node'], a['host'], a['host'], mode='backup')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                jsn = delete_node_data(a['host'], a['node'])
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                jsn = setup_node_directory(a['host'], a['node'], datadir_only = True)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                if use_ssl:
                    jsn = transfer_ssl_certificates(a['node'], a['host'], a['host'], mode='restore')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                jsn = start_node(a['host'], a['node'], in_replicaset=False)
                if jsn['stderr'] != '' and f"server running with dbpath:" not in jsn['stderr'].lower():
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                jsn = restart_node(a['host'], a['node'])
                if jsn['stderr'] != '' and f"server running with dbpath:" not in jsn['stderr'].lower():
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                error = using_primary_add_node_to_replicaset(primary['host'], primary['node'], a['host'], a['node'])
                if error != '':
                    last_known_err = error
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")

        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    try:
       with get_postgres_connection() as conn:
           with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true' WHERE cluster_id=%s """, (Severity.OK.value, Status.OK.value, target_version, primary['cluster_id']))
                conn.commit()
    except Exception as error:
       step = 'null'
       status = Severity.NOT_OK.value
       duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
       logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
       return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stepdown_mongo_primary_task(data, task_id):
    """a task to step down a primary node in a replica set"""

    task = 'step down node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['primary_host']
        node = json.loads(data['primary_data'])
        # list of secondaries/arbiters [{'id':, 'name':}, {}, ...]
        secondary_nodes = data['secondary_data']
        has_mongos = data['has_mongos']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stepping down'
    duration_step_start = time.perf_counter()

    error = step_down_node(host, node)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'searching for new primary'
    duration_step_start = time.perf_counter()

    (is_error, result) = find_replicaset_primary(host, node)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
        return

    error = update_cluster_roles(node, result, secondary_nodes, move_databases = not has_mongos)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def update_cluster_roles(old_primary, primary_name, secondary_nodes, move_databases = False) -> str:
    """link secondary/arbiters to new primary in app db"""

    if not primary_name:
        return 'cannot find primary'

    for n in secondary_nodes:
        if n.get('host'):
            n = n['node']

        if n['name'] == primary_name:
            try:
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET status = %s, ex_params = ex_params || '{"mongo_master_id":%s}' WHERE ex_params->>'replicaset_name'=%s""", (Status.OK.value, n['id'], old_primary['ex_params'].get('replicaset_name')))
                        cur.execute("""UPDATE "clusterCore_c123_node" SET cluster_role=%s WHERE  ex_params->>'replicaset_name'=%s AND cluster_role!=%s AND cluster_role!=%s""", (ClusterRoles.STANDBY.value, old_primary['ex_params'].get('replicaset_name'), ClusterRoles.ARBITER.value, ClusterRoles.BALANCER.value))
                        cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"mongo_master_id":%s}', cluster_role=%s WHERE id=%s""", (n['id'], ClusterRoles.PRIMARY.value, n['id']))
                        if move_databases:
                            cur.execute("""UPDATE "clusterCore_c123_database" SET node_id = %s WHERE node_id=%s""", (n["id"], old_primary['id']))
                        conn.commit()

                return ''
            except Exception as error:
                return str(error)

#############################
###   replicaset tasks    ###
#############################

def add_to_cfg(host, node, config_rs = False, shard_rs = False):
    """add the replicaset name to the node cfg file"""

    try:
        replicaset_name = node['ex_params'].get("replicaset_name")

        config_content = f"""
replication:
    replSetName: {replicaset_name}
"""
        if config_rs: 
            config_content = f"""{config_content}
sharding:
    clusterRole: configsvr
"""       
        
        if shard_rs:
            config_content = f"""{config_content}
sharding:
    clusterRole: shardsvr
"""
            
        data_path = node['dir_123Cluster']
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                pass

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        # remove replication key if it exists in config file
        remove_from_cfg(host, node)
        # add replica set to mongod.cfg file
        remote_command(host, f"echo '{config_content}' | sudo tee -a '{data_path}/conf/mongod.cfg'", True)
    except Exception as e:
        logging.error(f"add_to_cfg: host: {host}\nnode: {node}\nerror: {str(e)}")

def restart_node(host, node, secondary_nodes = None, in_replicaset = True, move_databases = False) -> dict:
    """restart node (if possible - in replica set)"""

    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        mongo_version = float(node['version'])
        data_path = node['dir_123Cluster']
        replicaset_name = node['ex_params'].get('replicaset_name')
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        mongo_bin = find_mongo_bin(host, mongo_version)
        # --noauth
        commands = [
            f"sudo {mongo_bin}/mongod --config '{data_path}/conf/mongod.cfg' --port '{node['port']}' --dbpath '{data_path}/data' --bind_ip '{host['name']}' --shutdown",
            f"""sudo {mongo_bin}/mongod --fork --config '{data_path}/conf/mongod.cfg' --port '{node['port']}' --dbpath '{data_path}/data' --bind_ip '{host['name']}' {f"--replSet '{replicaset_name}'" if in_replicaset else ''}"""
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                if not ('--shutdown' in command and any(err in jsn['stderr'].lower() for err in ["server running with dbpath", "does not appear to be running", "invalid process id"])):
                    return jsn

        if secondary_nodes:
            (is_error, result) = find_replicaset_primary(host, node)
            if is_error:
                jsn['stderr'] = result
                return jsn

            error = update_cluster_roles(node, result, secondary_nodes, move_databases = move_databases)
            if error != '':
                jsn['stderr'] = error
                return jsn

        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def add_first_node_to_replicaset(host, node, config_rs = False) -> str:
    """add the first node to the replicaset"""

    try:
        replicaset_name = node["ex_params"].get("replicaset_name")
        error = ''
        # connect directly to node
        client = create_mongo_connection(node, host['name'], ReadPreference.PRIMARY_PREFERRED.value, direct=True, no_auth=config_rs, no_ssl=config_rs)
        if isinstance(client, tuple):
            return str(client[1])
        
        with client:
            config = {
                '_id': replicaset_name,
                'members': [{
                    '_id': 0,
                    'host': f"{host['name'].strip()}:{node['port'].strip()}"
                }]
            }
            if config_rs:
                config['configsvr'] = True

            try:
                client.admin.command("replSetInitiate", config)
            except Exception as initiate_err:
                error, = initiate_err.args
                error = str(error)
                if 'already initialized' in error.lower():
                    # primary are in OTHER state or got automatically misconfigured,set right priority manually
                    conf = client.admin.command({'replSetGetConfig': 1})["config"]
                    conf["members"] = config["members"]
                    conf["members"][0]["priority"] = 1
                    conf["members"][0]["votes"] = 1
                    conf["version"] += 1 #bump config version
                    client.admin.command({"replSetReconfig": conf})
                    error = ''
                else:
                    return error

            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role =%s, ex_params = ex_params || '{"mongo_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.PRIMARY.value, node['id'], node['id'],))
                    conn.commit()

            return error
    except Exception as e:
        error, = e.args
        return str(error)

def add_node_to_rs_task(data, task_id):
    """
    add a node (arbiter or voting/non-voting secondary) to a replica set
    parameters are the ids of involved rs_id: the replicaset, primary_id:
    the primary node of the replicaset, node_id: the node that is added
    """

    task = 'add node to replicaset task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        primary_host = json.loads(data['host_primary'])
        secondary_host = json.loads(data['host_standby'])
        app_host = json.loads(data['app_host'])
        create_run_linux = data['create_run_linux']
        # repository = network or local
        repository = data['repository']
        secondary = json.loads(data['node_data'])
        primary = json.loads(data['primary_data'])
        use_ssl = primary['ex_params'].get('ssl')
        rs_name_exists = data['rs_name_exists']
        mongo_version = float(primary['version'])
        org_id = secondary['org_id']
        proxmox_container = secondary['ex_params'].get('proxmox')
        host_exists = True

        distributor, release = check_os_version(secondary_host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = secondary['ex_params'].get('distributor')
                release = secondary['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(secondary['id'])
                return

        version_exists = check_mongo_version_exists(secondary_host, mongo_version, distributor)

        # left intact, since we may use tilda paths in the future
        datadir_primary  =  primary['dir_123Cluster'].strip()
        datadir_secondary = secondary['dir_123Cluster'].strip()
        if datadir_primary.startswith('~'):
            jsn = find_tilda(primary_host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(secondary['id'])
                return

            tilda_path = jsn['stdout']
            datadir_primary = datadir_primary.replace('~', tilda_path)

        if datadir_secondary.startswith('~'):
            jsn = find_tilda(secondary_host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(secondary['id'])
                return

            tilda_path = jsn['stdout']
            datadir_secondary = datadir_secondary.replace('~', tilda_path)

        source_keyfile = f'{datadir_primary}/conf/key.file'
        target_keyfile = f'{datadir_secondary}/conf/key.file'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, secondary_host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(secondary['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    if not version_exists:
        step = 'installing mongodb'
        duration_step_start = time.perf_counter()

        if Repository.LOCAL.value in repository:
            jsn = install_offline_remote_to_remote(distributor, release, secondary_host, mongo_version, DBType.MONGO.value, app_host)
        else:
            jsn = install_mongo_online(distributor, release, secondary_host, mongo_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(secondary['id'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating node'
    duration_step_start = time.perf_counter()

    jsn = setup_node_directory(secondary_host, secondary)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(secondary['id'])
        return

    # all nodes of replica set must share keyfile of primary (for better failsafe - right after node directory creation)
    if primary_host['name'].strip() == secondary_host['name'].strip():
        try:
            jsn = json.loads(remote_command(primary_host, f"sudo rsync -a '{source_keyfile}' '{target_keyfile}'", True), strict = False)
            if jsn['stderr'] != '' or jsn['stdout'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(secondary['id'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    else:
        jsn = json.loads(remote_command(primary_host, f"sudo cat '{source_keyfile}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(secondary['id'])
            return

        keyfile_content = jsn['stdout']
        jsn = json.loads(remote_command(secondary_host, f"echo '{keyfile_content}' | sudo tee '{target_keyfile}' && sudo chmod 400 '{target_keyfile}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(secondary['id'])
            return

    if use_ssl:
        secondary_ssl_root = f"{datadir_secondary}/data"
        jsn = transfer_ssl_certificates(primary, primary_host, secondary_host, secondary_ssl_root)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        jsn = inject_ssl_config(secondary_host, datadir_secondary)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        jsn = write_ssl_to_file(secondary_host, secondary_ssl_root, secondary['port'])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    # for first boot up - any node should be started without --replSet
    jsn = start_node(secondary_host, secondary, in_replicaset=False)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(secondary['id'])
        return

    #if primary are standalone (unlikely to be execute, since we creating primary already in replicaset)
    if not rs_name_exists:
        add_to_cfg(primary_host, primary)

        jsn = restart_node(primary_host, primary)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(secondary['id'])
            return

        error = add_first_node_to_replicaset(primary_host, primary)
        if error != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            update_failed_object(secondary['id'])
            return

        if mongo_version >= 4.4:
            try:
                client = create_mongo_connection(primary, primary_host['name'])
                if isinstance(client, tuple):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {primary['name']}",str(client[1]),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {primary['name']}",str(client[1]),1)
                    return

                # connection for primary
                with client:
                    client.admin.command(({"setDefaultRWConcern" : 1, "defaultWriteConcern" : {"w" : 1}}))
            except Exception as e:
                error, = e.args
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

    # if node initialized correctly - restart it with --replSet
    # (mandatory for initial join of any voting nodes, for non-voting - may be skipped)
    add_to_cfg(secondary_host, secondary)

    jsn = restart_node(secondary_host, secondary)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(secondary['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'adding node to replicaset'
    duration_step_start = time.perf_counter()

    error = using_primary_add_node_to_replicaset(primary_host, primary, secondary_host, secondary)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(secondary_host, f"sudo du -sb '{secondary['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        remote_command(secondary_host, f"""echo "{get_log_rotate_config(secondary['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{secondary['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(secondary_host), 'node_name': secondary['name']}, 'task': 'log_rotate_task', 'monitor_name': None,  'org_id': secondary['org_id'],  'target_type': secondary['type'], 'target_id': secondary['id'], 'created_by': created_by}
        insert_task(params)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, secondary['version'], node_size, secondary['id'],))
                if primary['ex_params'].get('config_replicaset_name'):
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"severity":\"%s\"}' WHERE id=%s""" % (Severity.OK.value, primary['cluster_id'],))
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"status":\"\"}' WHERE id=%s""", (primary['cluster_id'],))
                
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,secondary['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,secondary_host['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
            
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)


    # secondary_host['create_task'] = created_by
    # secondary_host['ssh_port'] = int(secondary_host['ssh_port'])

    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': secondary_host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': secondary_host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as err:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)


def add_node_to_shard_task(data, task_id):
    """
    add a node (voting/non-voting secondary, arbiters are not supported in a sharded cluster topology) to a shard replicaset of sharded cluster
    parameters are the ids of involved rs_id: the replicaset name, primary_id:
    the primary node of the shard replicaset, node_id: the node that is added
    """

    task = 'add node to shard task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        primary_host = json.loads(data['host_primary'])
        secondary_host = json.loads(data['host_standby'])
        app_host = json.loads(data['app_host'])
        create_run_linux = data['create_run_linux']
        # repository = network or local
        repository = data['repository']
        secondary = json.loads(data['node_data'])
        primary = json.loads(data['primary_data'])
        use_ssl = primary['ex_params'].get('ssl')
        rs_name_exists = data['rs_name_exists']
        mongo_version = float(primary['version'])
        org_id = secondary['org_id']
        

        distributor, release = check_os_version(secondary_host)
        if OperatingSystem.INVALID.value in distributor:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,release['stdout'],release['stderr'],release['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,release['stdout'],release['stderr'],release['excode'])
            update_failed_object(secondary['id'])
            return

        version_exists = check_mongo_version_exists(secondary_host, mongo_version, distributor)

        # left intact, since we may use tilda paths in the future
        datadir_primary  =  primary['dir_123Cluster'].strip()
        datadir_secondary = secondary['dir_123Cluster'].strip()
        if datadir_primary.startswith('~'):
            jsn = find_tilda(primary_host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(secondary['id'])
                return

            tilda_path = jsn['stdout']
            datadir_primary = datadir_primary.replace('~', tilda_path)

        if datadir_secondary.startswith('~'):
            jsn = find_tilda(secondary_host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(secondary['id'])
                return

            tilda_path = jsn['stdout']
            datadir_secondary = datadir_secondary.replace('~', tilda_path)

        source_keyfile = f'{datadir_primary}/conf/key.file'
        target_keyfile = f'{datadir_secondary}/conf/key.file'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not version_exists:
        step = 'installing mongodb'
        duration_step_start = time.perf_counter()

        if Repository.LOCAL.value in repository:
            jsn = install_offline_remote_to_remote(distributor, release, secondary_host, mongo_version, DBType.MONGO.value, app_host)
        else:
            jsn = install_mongo_online(distributor, release, secondary_host, mongo_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(secondary['id'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating node'
    duration_step_start = time.perf_counter()

    jsn = setup_node_directory(secondary_host, secondary)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(secondary['id'])
        return

    # all nodes of replica set must share keyfile of primary (for better failsafe - right after node directory creation)
    if primary_host['name'].strip() == secondary_host['name'].strip():
        try:
            jsn = json.loads(remote_command(primary_host, f"sudo rsync -a '{source_keyfile}' '{target_keyfile}'", True), strict = False)
            if jsn['stderr'] != '' or jsn['stdout'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(secondary['id'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    else:
        jsn = json.loads(remote_command(primary_host, f"sudo cat '{source_keyfile}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(secondary['id'])
            return

        keyfile_content = jsn['stdout']
        jsn = json.loads(remote_command(secondary_host, f"echo '{keyfile_content}' | sudo tee '{target_keyfile}' && sudo chmod 400 '{target_keyfile}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(secondary['id'])
            return

    if use_ssl:
        secondary_ssl_root = f"{datadir_secondary}/data"
        jsn = transfer_ssl_certificates(primary, primary_host, secondary_host, secondary_ssl_root)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        jsn = inject_ssl_config(secondary_host, datadir_secondary)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        jsn = write_ssl_to_file(secondary_host, secondary_ssl_root, secondary['port'])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    # for first boot up - any node should be started without --replSet
    jsn = start_node(secondary_host, secondary, in_replicaset=False)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(secondary['id'])
        return

    #if primary are standalone (unlikely to be execute, since we creating primary already in replicaset)
    if not rs_name_exists:
        add_to_cfg(primary_host, primary)

        jsn = restart_node(primary_host, primary)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(secondary['id'])
            return

        error = add_first_node_to_replicaset(primary_host, primary)
        if error != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            update_failed_object(secondary['id'])
            return

        if mongo_version >= 4.4:
            try:
                client = create_mongo_connection(primary, primary_host['name'])
                if isinstance(client, tuple):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {primary['name']}",str(client[1]),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {primary['name']}",str(client[1]),1)
                    return

                # connection for primary
                with client:
                    client.admin.command(({"setDefaultRWConcern" : 1, "defaultWriteConcern" : {"w" : 1}}))
            except Exception as e:
                error, = e.args
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

    # if node initialized correctly - restart it with --replSet
    # (mandatory for initial join of any voting nodes, for non-voting - may be skipped)
    add_to_cfg(secondary_host, secondary, shard_rs=True)

    jsn = restart_node(secondary_host, secondary)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(secondary['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'adding node to replicaset'
    duration_step_start = time.perf_counter()

    error = using_primary_add_node_to_replicaset(primary_host, primary, secondary_host, secondary)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(secondary_host, f"sudo du -sb '{secondary['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024

        remote_command(secondary_host, f"""echo "{get_log_rotate_config(secondary['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{secondary['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(secondary_host), 'node_name': secondary['name']}, 'task': 'log_rotate_task', 'monitor_name': None,  'org_id': secondary['org_id'],  'target_type': secondary['type'], 'target_id': secondary['id'], 'created_by': created_by}
        insert_task(params)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, secondary['version'], node_size, secondary['id'],))
                if primary['ex_params'].get('config_replicaset_name'):
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"severity":\"%s\"}' WHERE id=%s""" % (Severity.OK.value, primary['cluster_id'],))
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"status":\"\"}' WHERE id=%s""", (primary['cluster_id'],))
                
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,secondary['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,secondary_host['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
                
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)


    # secondary_host['create_task'] = created_by
    # secondary_host['ssh_port'] = int(secondary_host['ssh_port'])

    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': secondary_host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': secondary_host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as err:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def using_primary_add_node_to_replicaset(primary_host, primary, host, node) -> str:
    """
    connects to mongodb using pymongo using primary
    get the replicaset config
    append the node to config.members
    bump the config version and set it to the replicaset
    """

    try:
        voting = bool(node['ex_params'].get('voting'))
        is_arbiter = ClusterRoles.ARBITER.value in node['cluster_role']
        client = create_mongo_connection(primary, primary_host['name'])
        if isinstance(client, tuple):
            return str(client[1])
        
        with client:
            conf = client.admin.command({'replSetGetConfig': 1})['config']
            # strip is recommended when building whitespace-sensitive parameter
            node_name = f"{host['name'].strip()}:{node['port'].strip()}"
            exists = any(i['host'] == node_name for i in conf['members'])
            if exists:
                # if already exists, thats fine, error - empty.
                return ''

            member_ids = [member['_id'] for member in conf['members']]
            # better calculate voting nodes number from primary config itself, like this
            voting_nodes = [member['_id'] for member in conf['members'] if int(member['votes']) > 0]
            if len(voting_nodes) > 6 and is_arbiter:
                return 'Error: arbiter should be voting node, but replicaset already reached maximum quorum of 7 voting nodes'
            
            is_voting = voting == True and len(voting_nodes) < 7
            id = 1
            while id in member_ids and id <= 255:
                id += 1

            if id > 255:
                return 'Error: replicaset already reached maximum amount of 256 nodes'

            node_conf = {
                '_id': id,
                'host': node_name,
            }

            if is_arbiter:
                node_conf['priority'] = 0
                node_conf['votes'] = 1
                node_conf['arbiterOnly'] = True
                conf['members'].append(node_conf)
                conf['version'] += 1  # Bump the config version
                client.admin.command({'replSetReconfig': conf})
            else:
                node_conf['priority'] = 0
                node_conf['votes'] = 1 if is_voting else 0
                conf['members'].append(node_conf)
                conf['version'] += 1  # Bump the config version
                client.admin.command({'replSetReconfig': conf})
                
                if is_voting:
                    time.sleep(10)
                    # wait until secondary catch up with primary
                    status = client.admin.command('replSetGetStatus')
                    primary_optime = -1
                    secondary_optime = -1
                    for n in status['members']:
                        if int(n['_id']) == id:
                            secondary_optime = n["optimeDate"]
                           
                        if ClusterRoles.PRIMARY.value in n['stateStr']:
                            primary_optime = n["optimeDate"] 
                    
                    if primary_optime != -1 and secondary_optime != -1:
                         lag = (primary_optime - secondary_optime).total_seconds()
                         while lag > 0:
                            time.sleep(10)
                            
                            status = client.admin.command('replSetGetStatus')
                            for n in status['members']:
                                    if int(n['_id']) == id:
                                        secondary_optime = n["optimeDate"]
                                    
                                    if ClusterRoles.PRIMARY.value in n['stateStr']:
                                        primary_optime = n["optimeDate"]
                            
                            lag = (primary_optime - secondary_optime).total_seconds() 
                    
                    conf = client.admin.command({'replSetGetConfig': 1})['config']
                    #bump priority of voting secondary
                    for n in conf['members']:
                        if int(n['_id']) == id:
                            n['priority'] = 1
                    
                    conf['version'] += 1 # Bump the config version
                        
                    client.admin.command({'replSetReconfig': conf})

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role=%s, ex_params = ex_params || '{"mongo_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.ARBITER.value if ClusterRoles.ARBITER.value in node['cluster_role'] else ClusterRoles.STANDBY.value, primary['id'], node['id'],))
                conn.commit()

        return ''
    except Exception as e:
        error, = e.args
        return str(error)

def remove_node_from_rs_task(data, task_id):
    """
    remove a node from a replica set
    parameters are the serialized primary, node to be removed and logger object
    """
    # unused so far

    task = 'remove node from replicaset task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        primary_host = json.loads(data['primary_host'])
        primary = json.loads(data['primary_data'])
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'running commands'
    duration_step_start = time.perf_counter()

    error = remove_node_from_rs_using_primary(primary_host, primary, node)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    remove_from_cfg(host, node)

    jsn = restart_node(host, node, in_replicaset=False)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def remove_node_from_rs_using_primary(primary_host, primary, node_hostname, node_port) -> str:
    """removes the node from the replica set member list using replica set primary node"""
    # 1. replSetReconfig is the internal command behind rs.remove(),
    # so rs.remove() has the same re-election side-effect
    # 2 May stuck if node for removal is stuck,
    # because mongo waits infinitely for config apply, by default
    # 3. this better than rs.remove(), to complete removal if node stuck - can provide force,
    # but since hard to tell if node stuck or just not accept connections - better manually edit function when needed.

    try:
        client = create_mongo_connection(primary, primary_host['name'])
        if isinstance(client, tuple):
            return str(client[1])

        with client:
            # get the config
            conf = client.admin.command({'replSetGetConfig': 1})
            # use same name as for node adding
            node_name = f"{node_hostname.strip()}:{node_port.strip()}"
            # exclude node from replica set members
            filtered_members = [obj for obj in conf['config']['members'] if obj['host'].strip() != node_name]
            # set the new member list
            conf['config']['members'] = filtered_members
            # Bump the config version
            conf['config']['version'] += 1
            # apply the changes
            client.admin.command({'replSetReconfig': conf['config']})
            return ''
    except Exception as e:
        error, = e.args
        return str(error)

def remove_from_cfg(host, node):
    """removes the replicaset from a mongo node config"""

    try:
        data_path = node['dir_123Cluster']
        if data_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                pass

            tilda_path = jsn['stdout']
            data_path = data_path.replace('~', tilda_path)

        conf_path = f"{data_path}/conf/mongod.cfg"
        # remove replication from mongod.cfg file
        remote_command(host, f"sudo sed -i '/replication:/d' '{conf_path}'", True)
        remote_command(host, f"sudo sed -i '/replSetName:/d' '{conf_path}'", True)
        remote_command(host, f"sudo sed -i '/sharding:/d' '{conf_path}'", True)
        remote_command(host, f"sudo sed -i '/clusterRole:/d' '{conf_path}'", True)
    except Exception as e:
        logging.error(f"remove_from_cfg: host: {host}\nnode: {node}\nerror: {str(e)}")

##############################
#    shard cluster task      #
##############################
#TODO When to do?
def initialte_shard_cluster_task(self, config_id, mongos_id, tpl_id):
    """
    initiates a shard cluster
    parameters:
    1. config server object id
    2. mongos object id
    3. task progress logger object id

    creating a shard cluster steps:
    1. stop the mongo process
    2. create the config server folders
    3. create the config server cfg file
    4. set execution permissions for the service
    5. start the config server service
    6. start the mongos service
    """

    # query objects from application backend
    # config = ConfigServer.objects.get(pk=config_id)
    # mongos = Mongos.objects.get(pk=mongos_id)
    # tpl = TaskProgressLog.objects.get(pk=tpl_id)

    # # serialize objects for sub tasks
    # config_data = object_to_json(config)
    # mongos_data = object_to_json(mongos)
    # # tpl_data = object_to_json(tpl)

    # # 1. stop the mongo process
    # stop_task = stop_mongo_process.si(

    #     config.host, config.user, config.pwd, tpl_id)

    # # 2. create the config server folders
    # config_dir_task = create_mongo_folders(host, node)

    # # 3. create the config server cfg file
    # create_config_task = set_mongo_config_file(host, node)

    # # 4. set execution permissions for the service
    # permissions_task = set_mongo_permissions(host, node)

    # # 5. start the config server service
    # start_server_task = start_node(host, node)

    # # 6. start the mongos service
    # deploy_task = deploy_mongos.si(config_data, mongos_data, tpl_id)

def stop_mongo_process(host, node_port) -> dict:
    """stop mongo process before making changes"""

    # thats force kill process by known port, use in exceptional situations only!

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node_port}' | awk '{{print $7}}' | awk -F'/' '{{print $1}}' | head -n 1", True), strict=False)
        if jsn['stderr'] != '':
            return jsn

        port_pid = jsn['stdout']
        #mongo docs forbid using kill -9 on mongod processes
        # (if using kill -2, mongo gracefully stop database,
        # and if primary node and can stepdown - do this first)
        jsn = json.loads(remote_command(host, f"sudo kill -2 {port_pid}", True), strict=False)
        if jsn['stderr'] != '':
            return jsn

        # wait (same as we wait for stepdown, for safety)
        time.sleep(120)
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node_port}' | awk '{{print $7}}' | awk -F'/' '{{print $1}}' | head -n 1", True), strict=False)
        if jsn['stderr'] != '':
            return jsn

        port_pid = jsn['stdout']
        if port_pid:
            # dont care about mongodb docs anymore,
            # node stuck and could be killed only by kill -9
            jsn = json.loads(remote_command(host, f"sudo kill -9 {port_pid}", True), strict=False)
            if jsn['stderr'] != '':
                return jsn

        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

##############################
#       database task        #
##############################

def create_mongo_database_task(data, task_id):

    task = 'create mongo database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['mongocluster_data'])
        db = json.loads(data['db'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating database users'
    duration_step_start = time.perf_counter()

    try:
        roles = [{"role": "readWrite", "db": str(db['name'])}]
        owner_user = f"{str(db['name'])}_owner_user"
        user_write = f"{str(db['name'])}_write_user"
        user_read = f"{str(db['name'])}_read_user"
        app_users = [owner_user, user_read, user_write]
        owner_user_pwd = generate_complex_password()
        read_user_pwd = generate_complex_password()
        write_user_pwd = generate_complex_password()
        mongos_port = node['ex_params'].get('mongos_port')
        logs_output = f'USERNAME = {owner_user}\nPASSWORD = {owner_user_pwd}\nUSERNAME = {user_read}\nPASSWORD = {read_user_pwd}\nUSERNAME = {user_write}\nPASSWORD = {write_user_pwd}'
        # connection for primary
        client = create_mongo_connection(node, host['name'], mongos_port=mongos_port)
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            return

        with client:
            db_obj = client[db['name']]
            # both as write test, and because mongodump ignore empty databases
            # use collection named "cluster123"
            col = db_obj["cluster123"]
            # a document
            doc = {"col_name": "cluster123", "host": f"{host['name']}"}
            # insert a document to the collection
            col.insert_one(doc)
            db_users = db_obj.command('usersInfo')
            for i in db_users['users']:
                if i['user'] in app_users:
                    db_obj.command("dropUser", i['user'])
            # Create owner user
            db_obj.command("createUser", owner_user, pwd=owner_user_pwd, roles=[{"role": "dbOwner", "db": str(db['name'])}])
            # Create read user
            db_obj.command("createUser", user_read, pwd=read_user_pwd, roles=[{"role": "read", "db": str(db['name'])}])
            # Create dml user
            db_obj.command("createUser", user_write, pwd=write_user_pwd, roles=roles, writeConcern={"w": "majority"})
            # Modify user to disallow drop and alter operations
            db_obj.command("revokeRolesFromUser", user_write, roles=[{"role": "dbOwner", "db": str(db['name'])}], writeConcern={"w": "majority"})
            db_obj.command("grantRolesToUser", user_write, roles=roles, writeConcern={"w": "majority"})
            if mongos_port:
                client.admin.command({'enableSharding': str(db['name'])})
            size = client[str(db['name'])].command('dbstats')
            db_size = float(size['storageSize'])/1024

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,logs_output,None,0)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_size,db['id'],))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_mongo_database_task(data, task_id):

    task = 'drop database task'

    try:
        # data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        db = data['db_data']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'dropping database'
    duration_step_start = time.perf_counter()

    error = drop_mongo_db(host, node, db['name'])
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DROPPED.value, db['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_mongo_db(host, node, db_name) -> str:

    try:
        # connection for primary
        client = create_mongo_connection(node, host['name'], mongos_port=node['ex_params'].get('mongos_port'))
        if isinstance(client, tuple):
            return str(client[1])
        
        with client:
            client.drop_database(db_name)

        return ''
    except Exception as e:
        error, = e.args
        return str(error)

def mongodump_task(data, task_id):
    """a task to use backup with "mongodump" command"""
    """if db_name don't have name then give None to parameter/s"""

    task = 'mongodump task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        mongo_version = float(node['version'])
        dump = json.loads(data['dump_data'])
        compress = dump['ex_params'].get('compress')
        use_ssl = node['ex_params'].get('ssl')
        db_name = data.get('db_name')
        if db_name:
            db_id = data.get('db_id')

        backup_dir = dump['ex_params'].get('backup_path')
        if backup_dir.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_dir = backup_dir.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        # d=directory, f=file
        dir_or_file = 'd'
        jsn = check_local_directory_exists(backup_dir, dir_or_file)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        exists = jsn['stdout']
        if 'false' in exists:
            jsn = json.loads(local_command(f"sudo mkdir -p '{backup_dir}'", True), strict=False)
            if jsn['stderr'] != '':
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    mongo_bin = find_local_mongo_bin(mongo_version)

    date = datetime.now().strftime('%d%m%Y_%H%M%S')
    if compress:
        path = f'{backup_dir}/{dump["name"]}_{date}.gz'
    else:
        path = f'{backup_dir}/{dump["name"]}_{date}.dump'

    step = 'running dump command'
    duration_step_start = time.perf_counter()

    cert_root = f"{TASK_CERTIFICATE_FILES_DIR}/{host['name'].strip()}_{node['port']}"
    ca = f"{cert_root}/ca.crt"
    key = f"{cert_root}/ca.pem"
    ssl_params = f"--ssl --sslCAFile='{ca}' --sslPEMKeyFile='{key}' --tlsInsecure" if use_ssl else ''

    # write dump in the mongo archive file
    if db_name is not None:
        # for database - specify database name
        # --dumpDbUsersAndRoles - override the default of not including users to dump
        # /root/123cluster/mongo/bin/mongodump --host 176.9.226.3 --port 27011 --db db_name --out=/root/123cluster/mongo/backup/dump/27011 --dumpDbUsersAndRoles
        command = f"sudo {mongo_bin}/mongodump --host='{host['name']}' --port='{node['ex_params'].get('mongos_port') if ClusterRoles.BALANCER.value in node['cluster_role'] else node['port']}' --username='{node['database_user']}' --password='{node['database_pwd']}' {ssl_params} --authenticationDatabase='admin' --db='{db_name}' --archive='{path}' --dumpDbUsersAndRoles"
    else:
        # node dumps - includes users, includes oplog
        # /root/123cluster/mongo/bin/mongodump --host 176.9.226.3 --port 27011 --out=/root/123cluster/mongo/backup/dump/27011
        #TODO: --oplog causing problems on standalone node dump (could be deprecated, we dont use standalones)
        # check if it is standalone (may be deprecated, we dont use standalones)
        (is_error, result) = find_replicaset_primary(host, node)
        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
            return

        # whether to include oplog in node dump
        oplog = '' if 'standalone' in result or node['name'] not in result else '--oplog'
        command = f"sudo {mongo_bin}/mongodump --host='{host['name']}' --port='{node['port']}' --username='{node['database_user']}' --password='{node['database_pwd']}' {ssl_params} --authenticationDatabase='admin' --archive='{path}' {oplog}"

    if compress:
        # apply mongo built-in gzip compression to mongo archive file
        command = f"{command} --gzip"

    jsn = json.loads(local_command(command, True), strict=False)
    if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking the dump size'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(local_command(f"sudo du -b '{path}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        size = jsn['stdout'].strip()
        backup_size = float(size)/1024
        if backup_size < 0.01:
            backup_size = 0.01

        modified_date = get_date_modified_folder(folder_path=backup_dir)
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}', status='true' WHERE id=%s""" % (modified_date, dump['id']))
                conn.commit()
                if db_name is None:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                    conn.commit()
                else:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_id,))
                    conn.commit()

                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,dump['id']))
                conn.commit()
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{backup_path}}', '"{0}"', true) WHERE id={1}""".format(path,dump['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def mongorestore_task(data, task_id):
    """a task to use restore with "mongorestore" command"""

    task = 'mongorestore task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        use_ssl = node['ex_params'].get('ssl')
        mongo_version = float(node['version'])
        dump = json.loads(data['dump'])
        compress = dump['ex_params'].get('compress')
        backup_host = data['backup_host']
        backup_file = dump['ex_params'].get('backup_path')
        db_name = data.get('db_name')
        if db_name:
            db_id = data.get('db_id')

        if backup_file.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_file = backup_file.replace('~', tilda_path)

        dir_or_file = 'f'
        jsn = check_local_directory_exists(backup_file, dir_or_file)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        exists = jsn['stdout']
        if 'false' in exists:
            error = f'the file {backup_file} does not exist'
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying dump file to primary'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(backup_host, host, os.path.dirname(backup_file), [backup_file])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'running restore command'
    duration_step_start = time.perf_counter()

    try:
        # connection for primary
        client = create_mongo_connection(node, host['name'], ReadPreference.PRIMARY_PREFERRED.value)
        if isinstance(client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
            return

        with client:
            ping = client.admin.command('ping').get('ok', 0.0)
    except Exception as e:
        logging.error(f"{task}@{step}: {task_id}: {str(e)}")
        ping = 0.0

    if ping == 0.0:
        jsn = restart_node(host, node)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    mongo_bin = find_local_mongo_bin(mongo_version)

    cert_root_node = f"{TASK_CERTIFICATE_FILES_DIR}/{host['name'].strip()}_{node['port']}"
    ca_node = f"{cert_root_node}/ca.crt"
    key_node = f"{cert_root_node}/ca.pem"
    ssl_params = f"--ssl --sslCAFile='{ca_node}' --sslPEMKeyFile='{key_node}' --tlsInsecure" if use_ssl else ''
    # restore dump from mongo archive file
    if db_name is not None:
        # for database - specify exact name
        # --restoreDbUsersAndRoles - override the default of not restoring users
        command = f"sudo {mongo_bin}/mongorestore --host='{host['name']}' --port='{node['ex_params'].get('mongos_port') if ClusterRoles.BALANCER.value in node['cluster_role'] else node['port']}' --username='{node['database_user']}' --password='{node['database_pwd']}' {ssl_params} --authenticationDatabase='admin' --archive='{backup_file}' --nsInclude='{db_name}.*' --db='{db_name}' --restoreDbUsersAndRoles"
    else:
        # node dumps -  includes oplog, always restores users
        # command = f"{mongo_bin}/mongorestore --host={host['name']} --port={node['port']} --username={node['database_user']} --password={node['database_pwd']} --authenticationDatabase=admin --archive={backup_dir} --oplogReplay"
        # check if node is standalone (may be deprecated, since we discarded standalones)
        (is_error, result) = find_replicaset_primary(host, node)
        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
            return

        # --oplogReplay - restore oplog entries also (for restore on replica set nodes, i wonder if work on secondary)
        oplog = '' if 'standalone' in result or node['name'] not in result else '--oplogReplay'
        command = f"sudo {mongo_bin}/mongorestore --host='{host['name']}' --port='{node['port']}' --username='{node['database_user']}' --password='{node['database_pwd']}' {ssl_params} --authenticationDatabase=admin --archive='{backup_file}' {oplog}"

    if db_name:
        command = f"{command} --db='{db_name}'"

    if compress:
        command = f"{command} --gzip"

    try:

        jsn = json.loads(local_command(command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])

    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024

        if db_name:
            client = create_mongo_connection(node, host['name'], ReadPreference.PRIMARY_PREFERRED.value)
            if isinstance(client, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {node['name']}",str(client[1]),1)
                return

            with client:
                size = client[db_name].command('dbstats')
                db_size = float(size['storageSize'])/1024

        remote_command(host, f"sudo rm -rf '{backup_file}'", True)
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                if db_name:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, size=%s, ind_started='true', ind_deleted = 'false' WHERE id=%s""", (Severity.OK.value, Status.OK.value,db_size,db_id,))

                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'updating database list'
    duration_step_start = time.perf_counter()

    error = update_dbs_list(host, node, created_by)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def mongo_backup_restore_task(data, task_id):
    """restore node from physical backup"""

    task = 'mongo backup restore task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_primary = data['host_data']
        primary = json.loads(data['primary_data'])
        use_ssl = primary['ex_params'].get('ssl')
        mongo_version = float(primary['version'])
        backup = json.loads(data['backup'])
        compress = backup['ex_params'].get('compress')
        arbiters = data.get('arbiters_list')
        standbys = data.get('standbys_list')
        # restore_option = 'this'/'latest'/'custom'
        restore_option = data['restore_option'].strip().lower()
        pitr = RestoreOptions.THIS.value not in restore_option
        custom_pitr = RestoreOptions.CUSTOM.value in restore_option
        # date is epoch(str)
        date = data['date']
        backup_host = data['backup_host']
        schedule = backup['is_sched']
        has_oplog = False

        jsn = find_local_tilda()
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        backup_dir = backup['ex_params'].get('backup_path')
        if backup_dir.startswith('~'):
            tilda_path = jsn['stdout']
            backup_dir = backup_dir.replace('~', tilda_path)

        if schedule:
            if restore_option in [RestoreOptions.LATEST.value, RestoreOptions.CUSTOM.value]:
                if custom_pitr:
                    timezone_cmd = 'cat /etc/timezone'
                    jsn = json.loads(local_command(timezone_cmd, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                    try:
                        timezone = pytz.timezone(jsn['stdout'])
                        target_timestamp = datetime.strptime(data['date'], "%Y-%m-%d %H:%M:%S").astimezone(timezone).timestamp()
                    except Exception as e:
                        error = f'Invalid server timezone: {jsn["stdout"]} error: {str(e)}'
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                        return

                    # get all backup folders, sorted by modified time ascending
                    command = f"sudo find '{backup_dir}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | grep -v '/dumps_oplog' | sort -n"
                else:
                    # get the last modified backup folder
                    command = f"sudo find '{backup_dir}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | grep -v '/dumps_oplog' | sort -n | tail -1 | awk '{{print $2}}'"
            else:
                # get oldest modified backup folder ('this')
                command = f"sudo find '{backup_dir}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | grep -v '/dumps_oplog' | sort -n | head -1 | awk '{{print $2}}'"

            jsn = json.loads(local_command(command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            if jsn['stdout'] == '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                return
            else:
                if restore_option in [RestoreOptions.THIS.value, RestoreOptions.LATEST.value]:
                    backup_dir = jsn['stdout']
                else:
                    output_lines = [line.strip() for line in jsn['stdout'].split("\n") if line.strip()]
                    closest_subfolder = None
                    min_time_difference = positive_infinity

                    for line in output_lines:
                        subfolder_timestamp_str, subfolder = line.split(" ", 1)
                        subfolder_timestamp = int(subfolder_timestamp_str)
                        time_difference = abs(target_timestamp - subfolder_timestamp)
                        if time_difference < min_time_difference:
                            min_time_difference = time_difference
                            closest_subfolder = subfolder

                    if not closest_subfolder:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                        return

                    backup_dir = closest_subfolder

        backup_root = os.path.dirname(backup_dir)
        oplog_root = f'{backup_root}/dumps_oplog'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # d=directory, f=file
    dir_or_file = 'd'
    jsn = check_local_directory_exists(backup_dir, dir_or_file)
    if jsn['stderr'] != '':
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    exists = jsn['stdout']
    if 'false' in exists:
        error = f"the backup directory {backup_dir} does not exist"
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    if pitr:
        step = 'checking oplog path'
        duration_step_start = time.perf_counter()

        try:
            # extract all directory paths under oplog root folder
            command = f"sudo find '{oplog_root}' -mindepth 1 -maxdepth 1 -type d"
            jsn = json.loads(local_command(command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            dir_list = [i.strip() for i in jsn['stdout']]
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    if host_primary['name'] != backup_host['name']:
        step = 'copying node backup to node host'
        duration_step_start = time.perf_counter()

        transfer_dirs = [backup_dir]
        if has_oplog and pitr:
            transfer_dirs.append(oplog_root)

        jsn = transfer_files_rsync(backup_host, host_primary, backup_root, transfer_dirs)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping nodes of replica set'
    duration_step_start = time.perf_counter()

    last_known_err = None

    if arbiters and len(arbiters) > 0:
        arbiters = [{"host": i["host"], "node": json.loads(i["node"])} for i in arbiters]
    else:
        arbiters = []

    if standbys and len(standbys) > 0:
        standbys = [{'host': i['host'], 'node': json.loads(i['node'])} for i in standbys]
    else:
        standbys = []

    for arbiter in arbiters:
        jsn = stop_node(arbiter['host'], arbiter['node'])
        if jsn['stderr'] != '':
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

    for standby in standbys:
        jsn = stop_node(standby['host'], standby['node'])
        if jsn['stderr'] != '':
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

    jsn = stop_node(host_primary, primary)
    if jsn['stderr'] != '':
        last_known_err = jsn
        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'replacing primary data with backup folder'
    duration_step_start = time.perf_counter()

    if compress:
        # subcommand for find latest tar.gz to unpack (compressed restore only)
        filename_cmd = f"sudo ls -lt --file-type '{backup_dir}' | grep '\.tar.gz$' | head -n 1 | awk '{{print $9}}'"
        jsn = json.loads(remote_command(host_primary, filename_cmd, True), strict = False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        archive_file = jsn['stdout']

    if use_ssl:
        jsn = transfer_ssl_certificates(primary, host_primary, host_primary, mode='backup')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    # what can be done by functions - should be done by functions
    jsn = delete_node_data(host_primary, primary)
    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in ['no such', 'no found']):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    # always preserve permissions on file transfer
    mv_cmd = f"sudo tar xzfC '{backup_dir}/{archive_file}' '{primary['dir_123Cluster']}'" if compress else f"sudo rsync -a --recursive '{backup_dir}/data/' '{primary['dir_123Cluster']}/data/'"

    commands = [
        mv_cmd,
        f"sudo chown -R $(whoami) '{primary['dir_123Cluster']}'"
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host_primary, command , True), strict=False)
            if jsn['stderr'] != '' or ('rsync' in command and jsn['stdout'] != ''):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                if backup_host['name'] != host_primary['name']:
                    remote_command(host_primary, f"sudo rm -rf '{backup_dir}'", True)
                    if schedule:
                        remote_command(host_primary, f"sudo rm -rf '{oplog_root}'", True)

                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    if use_ssl:
        jsn = transfer_ssl_certificates(primary, host_primary, host_primary, mode='restore')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'resetting the replica set'
    duration_step_start = time.perf_counter()

    remove_from_cfg(host_primary, primary)

    jsn = start_node(host_primary, primary)
    if jsn['stderr'] != '' and "server running with dbpath:" not in jsn['stderr'].lower():
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        if backup_host['name'] != host_primary['name']:
            remote_command(host_primary, f"sudo rm -rf '{backup_dir}'", True)
        return

    error = drop_mongo_db(host_primary, primary, 'local')
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        if backup_host['name'] != host_primary['name']:
            remote_command(host_primary, f"sudo rm -rf '{backup_dir}'", True)
        return

    add_to_cfg(host_primary, primary)

    jsn = restart_node(host_primary, primary)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        if backup_host['name'] != host_primary['name']:
            remote_command(host_primary, f"sudo rm -rf '{backup_dir}'", True)
        return

    # wait until replica set config discovery completes before adding nodes (no more than 10 seconds should be)
    time.sleep(10)

    error = add_first_node_to_replicaset(host_primary, primary)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        if backup_host['name'] != host_primary['name']:
            remote_command(host_primary, f"sudo rm -rf '{backup_dir}'", True)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if has_oplog and pitr:
        step = 'running mongorestore command'
        duration_step_start = time.perf_counter()

        cert_root_primary = f"{primary['dir_123Cluster'].strip()}/data"
        ca_primary = f"{cert_root_primary}/ca.crt"
        key_primary = f"{cert_root_primary}/ca.pem"
        ssl_params = f"--ssl --sslCAFile='{ca_primary}' --sslPEMKeyFile='{key_primary}' --tlsInsecure" if use_ssl else ''
        
        mongo_bin = find_mongo_bin(mongo_version)
        # replay oplog from oplog dump folders, one by one
        for dir in dir_list:
            command = f"sudo {mongo_bin}/mongorestore --host='{host_primary['name']}' --port='{primary['port']}' --username='{primary['database_user']}' --password='{primary['database_pwd']}' {ssl_params} --authenticationDatabase='admin' --oplogReplay '{dir}'"

            if custom_pitr:
                # if time provided - restrict replay to point in time
                command = f'{command} --oplogLimit {str(datetime.strptime(date, "%Y-%m-%d %H:%M:%S").astimezone(pytz.UTC).timestamp()).split(".")[0]}'
            
            jsn = local_command(command, True) 

        status = Severity.WARNING.value if jsn['stderr'] != '' else Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        if status == Severity.NOT_OK.value:
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        else:
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting nodes of the replica set'
    duration_step_start = time.perf_counter()

    last_known_err = None
    # "writeable" nodes (ones having data - primary/secondary)
    # MUST be started first
    for standby in standbys:
        standby_host = standby['host']
        standby_node = standby['node']
        if use_ssl:
            jsn = transfer_ssl_certificates(standby_node, standby_host, standby_host, mode='backup')
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        jsn = delete_node_data(standby_host, standby_node)
        if jsn['stderr'] != '':
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        jsn = setup_node_directory(standby_host, standby_node, datadir_only=True)
        if jsn['stderr'] != '':
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        if use_ssl:
            jsn = transfer_ssl_certificates(standby_node, standby_host, standby_host, mode='restore')
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        jsn = start_node(standby_host, standby_node, in_replicaset=False)
        if jsn['stderr'] != '' and f"server running with dbpath:" not in jsn['stderr'].lower():
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        jsn = restart_node(standby_host, standby_node)
        if jsn['stderr'] != '' and f"server running with dbpath:" not in jsn['stderr'].lower():
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        error = using_primary_add_node_to_replicaset(host_primary, primary, standby_host, standby_node)
        if error != '':
            last_known_err = error
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

    for arbiter in arbiters:
        arbiter_host = arbiter['host']
        arbiter_node = arbiter['node']
        if use_ssl:
            jsn = transfer_ssl_certificates(arbiter_node, arbiter_host, arbiter_host, mode='backup')
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        jsn = delete_node_data(arbiter_host, arbiter_node)
        if jsn['stderr'] != '':
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        jsn = setup_node_directory(arbiter_host, arbiter_node, datadir_only=True)
        if jsn['stderr'] != '':
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        if use_ssl:
            jsn = transfer_ssl_certificates(arbiter_node, arbiter_host, arbiter_host, mode='restore')
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        jsn = start_node(arbiter_host, arbiter_node, in_replicaset=False)
        if jsn['stderr'] != '' and f"server running with dbpath:" not in jsn['stderr'].lower():
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        jsn = restart_node(arbiter_host, arbiter_node)
        if jsn['stderr'] != '' and f"server running with dbpath:" not in jsn['stderr'].lower():
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        error = using_primary_add_node_to_replicaset(host_primary, primary, arbiter_host, arbiter_node)
        if error != '':
            last_known_err = error
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

    if mongo_version >= 4.4:
        try:
            client = create_mongo_connection(primary, host_primary['name'])
            if isinstance(client, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to node {primary['name']}",str(client[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to node {primary['name']}",str(client[1]),1)
                return

            # connection for primary
            with client:
                info = client.admin.command('hello')
                rs = info.get('setName')
                if rs:
                    client.admin.command(({"setDefaultRWConcern" : 1, "defaultWriteConcern" : {"w" : 1}}))
        except Exception as e:
            last_known_err = str(e)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'] if isinstance(last_known_err, dict) else None,last_known_err['stderr'] if isinstance(last_known_err, dict) else last_known_err,last_known_err['excode'] if isinstance(last_known_err, dict) else 1)

    step = 'updating database lists'
    duration_step_start = time.perf_counter()

    error = update_dbs_list(host_primary, primary, created_by)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host_primary, f"sudo du -sb '{primary['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, primary['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def mongodump_oplog_task(data, task_id):
    """a task to use backup with "mongodump" command to dump oplog collection"""
    """just for node schedule backup"""

    task = 'mongodump oplog task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        primary = json.loads(data['primary_data'])
        use_ssl = primary['ex_params'].get('ssl')
        mongo_version = float(primary['version'])
        mongo_shell = 'mongo' if mongo_version <= 4.4 else 'mongosh'
        # host of backed up node
        host_primary = data['host_primary']
        primary_hostname = host_primary['name']
        backup_host = data['backup_host']
        dump = json.loads(data['backup_data'])
        backup_dir = dump['ex_params'].get('backup_path')
        if backup_dir.startswith('~'):
            jsn = find_tilda(host_primary)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_dir = backup_dir.replace('~', tilda_path)

        # oplog_path = dump['ex_params'].get('oplog_path')
        mongo_bin = find_mongo_bin(host_primary, mongo_version)
        oplog_time = dump['ex_params'].get('oplog_time')
        if oplog_time == '':
            oplog_time = 'Timestamp({ t: 0, i: 0 })'

        date = datetime.now().strftime('%d%m%Y_%H%M%S')
        path = f'{backup_dir}/dumps_oplog/{date}'

        # parsing oplog_time from output by legacy binary json format string
        match = re.match(r'Timestamp\(\s*\{?\s*t\s*:\s*(\d+)\s*,\s*i\s*:\s*(\d+)\s*\}?\s*\)', oplog_time)
        if match:
            t, i = match.group(1), match.group(2)
        # if not matched
        else:
            # parsing oplog_time from output by modern binary json format string
            match = re.match(r'Timestamp\(\s*(\d+)\s*,\s*(\d+)\s*\)', oplog_time)
            if match:
                t, i = match.group(1), match.group(2)
            # if not falls into one of two formats - output is incorrect
            else:
                error = f'Invalid BSON Timestamp string representation {oplog_time}'
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'running dump command'
    duration_step_start = time.perf_counter()

    cert_root = f"{TASK_CERTIFICATE_FILES_DIR}/{host_primary['name'].strip()}_{primary['port']}"
    ca = f"{cert_root}/ca.crt"
    key = f"{cert_root}/ca.pem"
    ssl_params_dump = f"--ssl --sslCAFile='{ca}' --sslPEMKeyFile='{key}'" if use_ssl else ''
    ssl_params_shell = f"--tls --tlsCAFile='{ca}' --tlsCertificateKeyFile='{key}'" if use_ssl else ''

    # dump only oplog of node
    command = f"sudo {mongo_bin}/mongodump --host='{primary_hostname}' --port='{primary['port']}' --username='{primary['database_user']}' --password='{primary['database_pwd']}' {ssl_params_dump} --authenticationDatabase='admin' --db='local' --collection='oplog.rs' --out='{path}' "
    # take only portion of oplog, newer than position of previous oplog dump
    query = """--query '{"ts":{"$gt":{"$timestamp":{"t":%s,"i":%s}}}}'""" % (t, i)
    command = command + query

    try:
        jsn = json.loads(local_command(f"sudo mkdir -p '{path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    try:

        jsn = json.loads(local_command(command, True), strict=False)
        if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
       
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    # command = "bsondump --quiet {}/local/oplog.rs.bson | awk 'END { print }'".format(oplog_path)
    # connection string
    command = "sudo {0}/{1} --host '{2}' --port '{3}' -u '{4}' -p '{5}' {6} --authenticationDatabase 'admin' ".format(mongo_bin, mongo_shell, primary_hostname, primary['port'], primary['database_user'], primary['database_pwd'], ssl_params_shell)
    # find the current position of oplog
    line = f"""--quiet --eval 'db.getSiblingDB("local").oplog.rs.find({{}}, {{"ts":1}}).sort({{"$natural":-1}}).limit(1).next().ts;'"""
    command = command + line

    try:
        jsn = json.loads(local_command(command, True), strict=False)
        if jsn['stderr'] != '' or (jsn['stdout'] != '' and 'exception' in jsn['stdout'].lower()):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    oplog_time = jsn['stdout'].strip()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{oplog_path}}', '"{0}"', true) WHERE id={1}""".format(path,dump['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{oplog_time}}', '"{0}"', true) WHERE id={1}""".format(oplog_time,dump['id']))
                conn.commit()

        remote_command(host_primary, f"sudo rm -rf '{path}'", True)
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def copy_mongo_database_task(data, task_id):
    """copy database from some instance into another"""

    task = 'copy mongo database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_source = json.loads(data['host_source'])
        host_target = json.loads(data['host_target'])
        node_source = json.loads(data['node_source'])
        node_target = json.loads(data['node_target'])
        source_version = float(node_source['version'])
        target_version = float(node_target['version'])
        db_source = json.loads(data['db_source'])
        db_target = json.loads(data['db_target'])
        # db_option = 'override' or 'copy'
        db_option = data['db_option']
        db_exists = data['db_exists']
        # copy_option = data['copy_option']
        copy_name = data.get('copy_name')
        new_db_name = copy_name if isinstance(copy_name, str) and len(copy_name.strip()) > 0 else db_target["name"]
        source_ssl = node_source['ex_params'].get('ssl')
        target_ssl = node_target['ex_params'].get('ssl')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if db_exists and db_option == 'override':
        if host_source['name'] != host_target['name'] or node_source['port'] != node_target['port']:

            step = 'dropping existing database'
            duration_step_start = time.perf_counter()

            error = drop_mongo_db(host_target, node_target, db_source['name'])
            if error != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'copying the database'
    duration_step_start = time.perf_counter()

    mongo_bin_source = find_local_mongo_bin(source_version)
    mongo_bin_target = find_local_mongo_bin(target_version)
        
    cert_root_source = f"{TASK_CERTIFICATE_FILES_DIR}/{host_source['name'].strip()}_{node_source['port']}"
    ca = f"{cert_root_source}/ca.crt"
    key = f"{cert_root_source}/ca.pem"
    ssl_params_source = f"--ssl --sslCAFile='{ca}' --sslPEMKeyFile='{key}'" if source_ssl else ''
    # see mongodump_task
    dump_command = f"sudo {mongo_bin_source}/mongodump --host='{host_source['name']}' --port='{node_source['port']}' --username='{node_source['database_user']}' --password='{node_source['database_pwd']}' {ssl_params_source} --authenticationDatabase='admin' --db='{db_source['name']}' --archive='{TASK_TMP_FILES_DIR}/{host_source['name']}_{node_source['port']}_{db_source['name']}.dump' --dumpDbUsersAndRoles"
    cert_root_target = f"{TASK_CERTIFICATE_FILES_DIR}/{host_target['name'].strip()}_{node_target['port']}"
    ca = f"{cert_root_target}/ca.crt"
    key = f"{cert_root_target}/ca.pem"
    ssl_params_target = f"--ssl --sslCAFile='{ca}' --sslPEMKeyFile='{key}'" if target_ssl else ''
    # mostly the same as in mongorestore_task
    # --nsInclude, --nsFrom from which database to copy collections, --nsTo - to which database copy collections
    restore_command = f"sudo {mongo_bin_target}/mongorestore --host='{host_target['name']}' --port='{node_target['port']}' --username='{node_target['database_user']}' --password='{node_target['database_pwd']}' {ssl_params_target} --authenticationDatabase='admin' --archive='{TASK_TMP_FILES_DIR}/{host_source['name']}_{node_source['port']}_{db_source['name']}.dump' --nsInclude='{db_source['name']}.*' --nsFrom='{db_source['name']}.*' --nsTo='{new_db_name}.*' --db='{new_db_name}' --restoreDbUsersAndRoles"
    # depending on how big source database may be, copy through piping may possibly crash due to memory overflow, but we never tested app on big data
    command = f"{dump_command} && {restore_command} && sudo rm -rf '{TASK_TMP_FILES_DIR}/{host_source['name']}_{node_source['port']}_{db_source['name']}.dump'"
    
    try:
        commands = [
            f"sudo mkdir -p '{TASK_TMP_FILES_DIR}'",
            command
        ]
        for command in commands:
            jsn = json.loads(local_command(command, True), strict=False)
            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

   

    try:
        jsn = json.loads(remote_command(host_target, f"sudo du -sb '{node_target['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node_target['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_target['id'],))
                if not db_exists:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_source['id'],))

            conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.WARNING.value if status == Severity.NOT_OK.value else Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.WARNING.value if status == Severity.NOT_OK.value else Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def update_dbs_list(host, node, created_by) -> str:
    ignore = ['admin', 'local', 'config']
    error = ''
    try:
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory = RealDictCursor) as cur:
                cur.execute("""SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""", (node['id'],))
                db_list_from_app = cur.fetchall()
                db_set_from_app = {db['name'].strip() for db in db_list_from_app}
                # connection for primary
                client = create_mongo_connection(node, host['name'], ReadPreference.PRIMARY_PREFERRED.value)
                if isinstance(client, tuple):
                    return str(client[1])
                
                with client:
                    db_list_from_restore = {i['name'].strip(): float(i['sizeOnDisk'])/1024 for i in client.list_databases() if i['name'].strip() not in ignore}
                    db_set_from_restore = set(db_list_from_restore.keys())
                    common_db = db_set_from_app.intersection(db_set_from_restore)
                    missing_db = db_set_from_restore - db_set_from_app
                    extra_db = db_set_from_app - db_set_from_restore
                    for db in common_db:
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, "size"=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, db_list_from_restore[db], next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                        conn.commit()

                    for db in extra_db:
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, Status.OK.value, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                        conn.commit()

                    for db in missing_db:
                        cur.execute("""INSERT INTO "clusterCore_c123_database" 
                        ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id) 
                        VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", (db, DBType.MONGO.value, db_list_from_restore[db], Severity.OK.value, node['database_user'], node['database_pwd'], created_by, created_by, node['id']))
                        conn.commit()

                    return error
    except Exception as error:
        return str(error)
